import numpy as np
import logging
import os
import unittest
import doctest
from pysnptools.distreader import DistReader


class DistGen(DistReader):
    """
    A :class:`.DistReader` that generates -- on the fly -- deterministic "random" distributions over SNP data..

    See :class:`.DistReader` for general examples of using DistReader.

       **Tip:** The 'gen' in `DistGen` stands for generate, because it generates random (genetic) data. The 'gen' in :class:`Bgen` stands for 'genetic'. *

    **Constructor:**
        :Parameters: * **seed** (*number*) -- The random seed that (along with *block_size*) determines the SNP values.
        :Parameters: * **iid_count** (*number*) --  the number of iids (number of individuals)
        :Parameters: * **sid_count** (*number*) --  the number of sids (number of SNPs)
        :Parameters: * **chrom_count** (*number*) --  the number of chromosomes to generate. Default is 22. Must be 22 or less.
        :Parameters: * **cache_file** (*string*) -- (default None) If provided, tells where to cache the common iid, sid, and pos information. Using it can save time.
        :Parameters: * **block_size** (*number* or None) -- (default None) Tells how many SNP to generate at once. The default picks a *block_size* such that *iid_count \* *block_size* is about 100,000.

        :Example:

        >>> from pysnptools.distreader import DistGen
        >>> #Prepare to generate data for 1000 individuals and 1,000,000 SNPs
        >>> dist_gen = DistGen(seed=332,iid_count=1000,sid_count=1000*1000)
        >>> print(dist_gen.iid_count,dist_gen.sid_count)
        1000 1000000
        >>> dist_data = dist_gen[:,200*1000:201*1000].read() #Generate data for all users and for SNPs 200K to 201K
        >>> print(dist_data.val[2,2], dist_data.iid_count, dist_data.sid_count)
        [0.2760024  0.19624834 0.52774925] 1000 1000

    """

    def __init__(
        self,
        seed,
        iid_count,
        sid_count,
        chrom_count=22,
        cache_file=None,
        block_size=None,
    ):  # LATER fails on 450000x1
        self._ran_once = False
        self._cache_file = cache_file

        self._seed = seed
        self._iid_count = iid_count
        self._sid_count = sid_count
        self._chrom_count = chrom_count
        self._block_size = block_size or max((100 * 1000) // max(1, iid_count), 1)

        if cache_file is not None:
            if not os.path.exists(cache_file):
                import pysnptools.util as pstutil

                pstutil.create_directory_if_necessary(cache_file)
                self._run_once()
                np.savez(
                    cache_file,
                    _row=self._row,
                    _col=self._col,
                    _col_property=self._col_property,
                )
            else:
                with np.load(cache_file, allow_pickle=True) as data:
                    self._row = data["_row"]
                    assert (
                        len(self._row) == iid_count
                    ), "The iid in the cache file has a different length than iid_count"
                    self._col = data["_col"]
                    assert (
                        len(self._col) == sid_count
                    ), "The sid in the cache file has a different length than sid_count"
                    self._col_property = data["_col_property"]
                    assert (
                        len(self._col_property) == sid_count
                    ), "The pos in the cache file has a different length than sid_count"
                    self._ran_once = True

    def __repr__(self):
        return "{0}(seed={1},iid_count={2},sid_count={3},chrom_count={4},block_size={5},cache_file={6})".format(
            self.__class__.__name__,
            self._seed,
            self._iid_count,
            self._sid_count,
            self._chrom_count,
            self._block_size,
            self._cache_file,
        )

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    _chrom_size = np.array(
        [
            263,
            255,
            214,
            203,
            194,
            183,
            171,
            155,
            145,
            144,
            144,
            143,
            114,
            109,
            106,
            98,
            92,
            85,
            67,
            72,
            50,
            56,
        ],
        dtype=np.int64,
    ) * int(
        1e6
    )  # The approximate size of human chromosomes in base pairs

    def _run_once(self):
        if self._ran_once:
            return
        self._ran_once = True
        self._row = np.array(
            [("0", "iid_{0}".format(i)) for i in range(self._iid_count)]
        )
        self._col = np.array(["sid_{0}".format(i) for i in range(self._sid_count)])
        self._col_property = np.zeros(
            ((self._sid_count), 3)
        )  # Must be zero (not empty) because will leave _col_property[:,1] unchanged

        chrom_total = DistGen._chrom_size[: self._chrom_count].sum()
        step = chrom_total // self._sid_count
        start = 0
        chrom_size_so_far = 0
        for chrom_index in range(self._chrom_count):
            chrom_size_so_far += DistGen._chrom_size[chrom_index]
            stop = chrom_size_so_far * self._sid_count // chrom_total
            self._col_property[start:stop, 0] = chrom_index + 1
            random_increment = np.random.RandomState(chrom_index ^ 2292).randint(step)
            self._col_property[start:stop, 2] = (
                np.arange(0, stop - start) * step + random_increment + 1
            )
            # print(chrom_index+1,start,stop,self._sid_count)
            start = stop

    def copyinputs(self, copier):
        self._run_once()
        copier.input(self._cache_file)

    # Most _read's support only indexlists or None, but this one supports Slices, too.
    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        self._run_once()
        import pysnptools.util as pstutil

        if order == "A":
            order = "F"
        dtype = np.dtype(dtype)

        row_index_count = (
            len(row_index_or_none) if row_index_or_none is not None else self._iid_count
        )  # turn to a count of the index positions e.g. all of them
        col_index = (
            col_index_or_none
            if col_index_or_none is not None
            else np.arange(self._sid_count)
        )  # turn to an array of index positions, e.g. 0,1,200,2200,10
        batch_index = (
            col_index // self._block_size
        )  # find the batch index of each index position, e.g. 0,0,0,2,0
        val = np.empty(
            (row_index_count, len(col_index), 3), order=order, dtype=dtype
        )  # allocate memory for result
        list_batch_index = list(set(batch_index))
        for ii, i in enumerate(
            list_batch_index
        ):  # for each distinct batch index, generate dists # !!!fix up snpgen this way, too with ii
            i = int(i)  # convert np.uint64 to python int to avoid uint*int->float
            # LATER logging.info("working on distgen batch {0} of {1}".format(ii,len(list_batch_index))) # !!!why does this produce messages like 'working on distgen batch 8 of 2'?
            start = i * self._block_size  # e.g. 0 (then 2000)
            stop = start + self._block_size  # e.g. 1000, then 3000
            batch_val = self._get_val(start, stop, dtype)  # generate whole batch
            a = (
                batch_index == i
            )  # e.g. [True,True,True,False,True], then [False,False,False,True,False]
            b = col_index[a] - start  # e.g.  0,1,200,10, then 200
            val[:, a, :] = (
                batch_val[:, b, :]
                if row_index_or_none is None
                else pstutil.sub_matrix(
                    batch_val, row_index_or_none, b, num_threads=num_threads
                )
            )

        return val

    def _get_dist(self):
        w = np.array([-0.6482249, -8.49790398])

        def dist_fit(x):
            log_y = w[0] * np.log(x) + w[1]
            return np.exp(log_y)

        x_sample = np.logspace(
            np.log10(0.1 / self._iid_count), np.log10(0.5), 100, base=10
        )  # discretize from 1e-7 (assuming iid_count=1e6) to .5, logarithmically
        y_sample = np.array(
            [dist_fit(x) for x in x_sample]
        )  # Find the relative weight of each point
        dist = y_sample / y_sample.sum()
        return x_sample, dist

    @staticmethod
    def _get_sid(sid_start, sid_stop):
        sid = ["sid_{0}".format(i) for i in range(sid_start, sid_stop)]
        return sid

    def _get_val(self, sid_start, sid_stop, dtype):
        missing_rate = 0.218
        block_size = sid_stop - sid_start

        np.random.seed(self._seed + sid_start)
        val = np.random.random((self.iid_count, block_size, 3))
        val /= val.sum(axis=2, keepdims=True)  # make probabilities sum to 1

        missing = np.random.rand(self.iid_count, block_size) < missing_rate
        val[missing, :] = np.nan

        return val


class TestDistGen(unittest.TestCase):
    def test1(self):
        logging.info("in TestDistGen test1")
        seed = 0
        distgen = DistGen(
            seed=seed, iid_count=1000, sid_count=1000 * 1000, block_size=1000
        )
        distdata = distgen[:, [0, 1, 200, 2200, 10]].read()
        distdata2 = distgen[:, [0, 1, 200, 2200, 10]].read()
        assert distdata.allclose(distdata2)

        from pysnptools.distreader import DistNpz

        ref = DistNpz(
            os.path.dirname(os.path.realpath(__file__))
            + "/../../tests/datasets/distgen.dist.npz"
        ).read()
        assert distdata.allclose(ref, equal_nan=True)

        cache_file = "tempdir/cache_file_test1.npz"
        os.remove(cache_file) if os.path.exists(cache_file) else None
        distgen3 = DistGen(
            seed=seed,
            iid_count=1000,
            sid_count=1000 * 1000,
            block_size=1000,
            cache_file=cache_file,
        )
        distdata3 = distgen3[::10, [0, 1, 200, 2200, 10]].read()
        assert distdata3.allclose(distdata2[::10, :].read())
        distgen4 = DistGen(
            seed=seed,
            iid_count=1000,
            sid_count=1000 * 1000,
            block_size=1000,
            cache_file=cache_file,
        )
        distdata4 = distgen4[::10, [0, 1, 200, 2200, 10]].read()
        assert distdata4.allclose(distdata2[::10, :].read())

    def test_doctest(self):
        import pysnptools.distreader.distgen as gen_mod

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(gen_mod)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistGen))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    result = doctest.testmod()
    assert result.failed == 0, "failed doc test: " + __file__



"""
Tools for reading and manipulating SNP distribution data. For each individual and SNP location, it gives three probabilities: P(AA),P(AB),P(BB),
where A and B are the alleles. The probabilities should sum to 1.0. Missing data is represented by three numpy.NaN's.
"""

from pysnptools.distreader.distreader import DistReader
from pysnptools.distreader.distdata import DistData
from pysnptools.distreader.distnpz import DistNpz
from pysnptools.distreader.disthdf5 import DistHdf5
from pysnptools.distreader.distmemmap import DistMemMap
from pysnptools.distreader.bgen import Bgen
from pysnptools.distreader.distgen import DistGen
from pysnptools.distreader._distmergesids import _DistMergeSIDs




"""Tools for reading and writing files, locally or across clusters.
"""

import logging
import os
import random
import re
import socket
import threading
import time


# see https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
def ip_address():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(("10.254.254.254", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


def ip_address_pid():
    """
    Return the ip address of this computer and process id of the executing process.
    """
    ip = ip_address()
    pid = os.getpid()
    tid = threading.current_thread().ident
    return "{0}.{1}.{2}".format(ip, pid, tid)


class _DibLib(object):
    def __init__(self, unique_name, dib_path, dir_path, dib_name):
        self.dib_path = dib_path
        self.dir_path = dir_path
        self.dibs_fn = "{0}_{1}.txt".format(dib_name, unique_name)
        self.dibs_pattern = re.compile(r"^{0}_.*\.txt$".format(dib_name))
        if dib_name == "dibs":  # could make two subclasses, but this seems simpler
            self.test = self.tree_copy
        elif dib_name == "azure_storage_dib":
            self.test = self.only_first_goes
        else:
            raise Exception(
                "Don't know what test goes with dib_name of '{0}'".format(dib_name)
            )

    def tree_copy(self, priority, dibs_time, try_index):
        copy_count = len(
            [f for f in self.dir_path.walk() if PeerToPeer._copy_main_pattern.match(f)]
        )  # count main + copies
        logging.info(
            "PeerToPeer priority is {0} and copy_count is {1}".format(
                priority, copy_count
            )
        )
        if priority + 1 <= copy_count * 2:
            return "go"
        else:
            return "wait"

    def only_first_goes(self, priority, dibs_time, try_index):
        logging.info("P2P repair priority is {0}".format(priority))
        if priority == 0 and try_index == 0:
            return "azure"  # If dibs first, then go
        if priority == 0:
            logging.info(
                "Priority was > 0, but there is now a p2p file to copy, so the wait is over"
            )
            return "fixed"
        for f in self.dir_path.walk():
            if (
                PeerToPeer._copy_main_pattern.match(f)
                and self.dir_path._simple_getmtime(f) > dibs_time
            ):
                logging.info(
                    "Priority > 0, but there is now a p2p file to copy, so the wait is over"
                )
                return "fixed"  # If dibs later, but there is a new "main" or "copy" file, then go
        else:
            return "wait"  # else, wait

    def ever(self):
        i = 0
        while True:
            yield i
            i += 1

    def wait_for_turn(self):
        self.dib_path.save(self.dibs_fn, "")
        time.sleep(
            1.0
        )  # Sleep one second to (hopefully) see dibs made at exactly the same time.
        dibs_time = self.dib_path.getmtime(self.dibs_fn)
        sleep_time = 5.0
        for try_index in self.ever():
            priority = self.get_priority(
                dibs_time
            )  # Count number of remaining dibs files that were created before mine. That is my priority.
            status = self.test(priority, dibs_time, try_index)
            if status != "wait":
                logging.info("I can go! (status='{0}')".format(status))
                return status
            logging.info("sleep for {0}".format(sleep_time))
            time.sleep(sleep_time)
            sleep_time = min(
                sleep_time * random.uniform(1.0, 1.2), 60
            )  # wait 0 to 10% longer, but no more than a minute

    def remove_dibs(self):
        if self.dib_path._simple_file_exists(self.dibs_fn):
            self.dib_path._simple_remove(self.dibs_fn)

    def get_priority(self, dibs_time):
        priority = 0
        sequence = (
            file for file in self.dib_path.walk() if self.dibs_pattern.match(file)
        )
        for other_dib in sequence:
            try:  # It might disappear out from under us
                other_time = self.dib_path.getmtime(other_dib)
            except Exception:
                other_time = None
            if (
                other_time is None
                or other_time < dibs_time
                or (other_time == dibs_time and other_dib < self.dibs_fn)
            ):
                priority += 1
        return priority


from pysnptools.util.filecache.filecache import FileCache  # noqa: F401
from pysnptools.util.filecache.hashdown import Hashdown  # noqa: F401
from pysnptools.util.filecache.localcache import LocalCache  # noqa: F401
from pysnptools.util.filecache.peertopeer import PeerToPeer  # noqa: F401

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import absolute_import
import sys
import re
from bisect import bisect_left
import logging
import numpy as np
import unittest
import doctest
import numbers
import io


class IntRangeSet(object):
    """
    A class for efficiently manipulating ranges of integers (including negatives and longs) using set operations such as :meth:`union`, :meth:`intersection`, and difference.

    The class differs from the built-in *set* class (and from Boolean numpy arrays) because it does not need to store every element in the set, only for every contiguous range of elements.
    It differs from other Python interval libraries (that we know of) by being specialized and optimized for integer elements.

    :Example:

    Here we take the union (operator "|") of two IntRangeSets:

    .. image:: example1.png

    >>> from pysnptools.util import IntRangeSet
    >>> a = IntRangeSet("100:500,501:1000") # a is the set of integers from 100 to 500 (exclusive) and 501 to 1000 (exclusive)
    >>> b = IntRangeSet("-20,400:600")      # b is the set of integers -20 and the range 400 to 600 (exclusive)
    >>> c = a | b                           # c is the union of a and b, namely -20 and 100 to 1000 (exclusive)
    >>> print(c)
    IntRangeSet('-20,100:1000')

    :Example:

    Suppose we want to find the intron regions of a gene but we are given only the transcription region and the exon regions.

    .. image:: example2.png

    >>> from pysnptools.util import IntRangeSet
    >>> line = "chr15   29370   37380   29370,32358,36715   30817,32561,37380"
    >>> chr,trans_start,trans_last,exon_starts,exon_lasts = line.split() # split the line on white space
    >>> trans_start = int(trans_start)
    >>> trans_stop = int(trans_last) + 1 # add one to convert the inclusive "last" value into a Pythonesque exclusive "stop" value
    >>> int_range_set = IntRangeSet((trans_start,trans_stop)) # creates a IntRangeSet from 29370 (inclusive) to 37381 (exclusive)
    >>> print(int_range_set) # print at any time to see the current value
    IntRangeSet('29370:37381')

    Parse the exon start and last lists from strings to lists of integers (converting 'last' to 'stop')

    >>> exon_starts = [int(start) for start in exon_starts.strip(",").split(',')]
    >>> exon_stops = [int(last)+1 for last in exon_lasts.strip(",").split(',')]
    >>> assert len(exon_starts) == len(exon_stops)

    Zip together the two lists to create an iterable of exon_start,exon_stop tuples. Then
    'set subtract' all these ranges from int_range_set

    >>> int_range_set -= zip(exon_starts,exon_stops)
    >>> print(int_range_set) # See what it looks like
    IntRangeSet('30818:32358,32562:36715')

    Create the desired output by iterating through each contiguous range of integers

    >>> for start, stop in int_range_set.ranges():
    ...    print("{0}\t{1}\t{2}".format(chr, start, stop-1))
    chr15   30818     32357
    chr15   32562     36714


    **Ranges Input**

    The input to the :class:`IntRangeSet` constructor and many of its methods is a *ranges input*.

    A *ranges input* is one of the following:

        * A comma-separated string of integers or integer ranges, e.g., ``'100:500,500:1000,2000'``
        * An integer or integer expression, e.g., ``3``
        * A tuple of two integers, *start* and *stop*, e.g., ``(2,8)``. *stop* is exclusive.
        * A slice with non-negative values, e.g. ``slice(2,8)``
        * A :class:`IntRangeSet` (or any class with a :meth:`ranges` method), e.g., ``IntRangeSet(3)``
        * A list or iterable (but not tuple) of *ranges inputs*, e.g., ``[1,6,7,(100,200)]``

        :Example:

        Strings:

        >>> a = IntRangeSet("100:500,500:1000,2000")
        >>> a = IntRangeSet('-10:-8,1,2,3:11')
        >>> a = IntRangeSet('')

        The ranges in an *ranges input* can overlap and can be in any order.

        >>> assert IntRangeSet("2000,100:1500,500:1000,2000") == IntRangeSet('100:1500,2000')

        Integers and Integer Expressions:

        >>> a = IntRangeSet(7)
        >>> a = IntRangeSet(151000000000) # longs are OK
        >>> a = IntRangeSet(2*3+1) # integer expressions are OK

        A tuple must have exactly two integers. They represent the *start* (inclusive) and *stop* (exclusive) integers in a range.

        >>> assert IntRangeSet((2,8)) == IntRangeSet('2:8') # check 'set equality'
        >>> assert 7 in IntRangeSet((2,8)) # The integer 7 is an element of the set
        >>> assert IntRangeSet((2,3)) == IntRangeSet(2)

        Lists and iterables of *ranges inputs*:

        >>> assert IntRangeSet([1,6,7,(100,200)]) == IntRangeSet('1,6:8,100:200')
        >>> assert IntRangeSet(range(0,100)) == IntRangeSet('0:100')
        >>> assert IntRangeSet([range(100,200),3,'1000:2000',[4,6],(20,30)]) == IntRangeSet('3:5,6,20:30,100:200,1000:2000')



        Some methods accept zero or more *ranges input* as their input. This is called a *ranges_inputs*. For example:

            >>> a = IntRangeSet() # zero ranges inputs
            >>> a = IntRangeSet(3) # one ranges input
            >>> a = IntRangeSet('3:11,5',14,100) # Three ranges inputs, a string and two integers.


    **Most Important Methods and Operators**

    ================================== ================================== ======================
    *Description*                       *Method or Operator*              *For Details*
    is a 'set equal' to b?             ``a == b``                         :meth:`__eq__`
    is a is superset of b?             ``b in a``,                        :meth:`__contains__`
                                       ``a.issuperset(b)``,
                                       ``a >= b``
    is a is subset of b?               ``a <= b``,                        :meth:`__le__`
                                       ``a.issubset(b)``
    remove *i* th smallest element(s)  ``del a[i]``                       :meth:`__delitem__`
    get *i* th smallest element(s)     ``a[i]``                           :meth:`__getitem__`
    is a not 'set equal' to b?         ``a != b``                         :meth:`__ne__`
    is a is proper superset of b?      ``a > b``                          :meth:`__gt__`
    is a is proper subset of b?        ``a < b``                          :meth:`__lt__`
    union b into a                     ``a |= b``,                        :meth:`add`
                                       ``a += b``,
                                       ``a.add(b,c)``,
                                       ``a.update(b,c)``
    union a and b                      ``a | b``,                         :meth:`__or__`
                                       ``a b``,
                                       ``a.union(b,c)``
    intersect b into a                 ``a &= b``,                        :meth:`__iand__`
                                       ``a.intersection_update(b,c)``,
    intersect a and b                  ``a & b``,                         :meth:`intersection`
                                       ``a.intersection(b,c)``
    subtract b from in a               ``a -= b``,                        :meth:`__isub__`
                                       ``a.difference_update(b)``,
                                       ``a.discard(b)``,
    remove b from a, error if missing  ``a.remove(b)``                    :meth:`remove`
    a 'set difference' b                ``a - b``,                        :meth:`__sub__`
                                        ``a.difference(b)``
    iterate integers in a, from low    ``for element in a:``              :meth:`__iter__`
    iterate integers in a, from high   ``for element in reverse(a):``     :meth:`__reversed__`
    symmetric difference into a        ``a ^= b``                         :meth:`__ixor__`
    a 'symmetric difference' b         ``a ^ b``                          :meth:`__xor__`
    count of integer elements in a     ``len(a)``                         :meth:`__len__`
    iterate ranges in a, from low      ``for start,stop in a.ranges():``  :meth:`ranges`
    count of ranges in a               ``a.ranges_len``                   :meth:`ranges_len`
    index of range containing b        ``a.ranges_index(b)``              :meth:`ranges_index`
    get *i* th smallest range          ``a.ranges_getitem(i)``            :meth:`ranges_getitem`
    a as a string                      ``str(a)``                         :meth:`__str__`
    remove all elements from a         ``a.clear()``                      :meth:`clear`
    copy a                             ``a.copy()``                       :meth:`copy`
    find index of b in a               ``a.index(b)``                     :meth:`index`
    are a and b disjoint?              ``a.isdisjoint(b)``                :meth:`isdisjoint`
    is a empty?                        ``a.isempty(b)``                   :meth:`isempty`
    smallest element of a              ``a.min()``                        :meth:`min`
    largest element of a               ``a.max()``                        :meth:`max`
    sum of elements in a               ``a.sum()``                        :meth:`sum`
    remove and return an element       ``a.pop()``                        :meth:`pop`
    ================================== ================================== ======================

    **Examples, Tips, and Warnings**

    The argument to a method and the right-hand side of an operator can be a *ranges input* rather than :class:`IntRangeSet`. For example,

        >>> big = IntRangeSet('-1000:2000')
        >>> print(big - '10:20,500') # We can subtract this string because it is a legal ranges input.
        IntRangeSet('-1000:10,20:500,501:2000')

        This even works for equality testing:

        >>> assert big - '10:21,500' == '-1000:10,21:500,501:2000'

        The Python *in* operator is backwards from other operators, so the left-hand side can be any *ranges input*.

        >>> assert 501 in big

    Like most other Python libraries you specify a range with an inclusive *start* integer and an exclusive *stop*.

        >>> print(7 in IntRangeSet('4:8')) # includes 7
        True
        >>> print(8 in IntRangeSet('4:8')) # excludes 8
        False
        >>> print(8 in IntRangeSet(range(4,7))) # also excludes 8
        False

    Be careful with the *ranges inputs* specified with tuples. Suppose we want a :class:`IntRangeSet` containing 4,5,6,7.
        >>> assert IntRangeSet((4,8)) == '4:8' #OK
        >>> assert IntRangeSet([(4,8)]) == '4:8' # List of a tuple is OK
        >>> assert IntRangeSet(4,8) == '4,8' #No. This is not a tuple. It is two integer inputs, so we get only 4 and 8
        >>> assert IntRangeSet([4,8]) == '4,8' #No. List of two integers, so get only 4 and 8
        >>> #Illegal: IntRangeSet((4,8,10)) # tuples must be pairs


    **Methods and Operators**
    """

    _rangeExpression = re.compile(r"^(?P<start>-?\d+)(\:(?P<stop>-?\d+))?$")

    def __init__(self, *ranges_inputs):
        """
        Create a :class:`IntRangeSet`.
        """
        if len(ranges_inputs) > 0 and isinstance(
            ranges_inputs[0], IntRangeSet
        ):  # Because we know self is empty, optimize for the case in which the first item is a IntRangeSet
            self._start_items = list(ranges_inputs[0]._start_items)
            self._start_to_length = dict(ranges_inputs[0]._start_to_length)
            ranges_inputs = ranges_inputs[1:]
        else:
            self._start_items = []
            self._start_to_length = {}
        self.add(*ranges_inputs)

    def add(self, *ranges_inputs):
        """
        Union zero or more ranges inputs into the current IntRangeSet.

        These are the same:

        * ``a |= b``
        * ``a += b``
        * ``a.add(b)``
        * ``a.update(b)``

        :Example:

        >>> a = IntRangeSet('0:5,6:10')
        >>> a |= 5
        >>> print(a)
        IntRangeSet('0:10')


        The 'add' and 'update' methods also support unioning multiple ranges inputs,

        :Example:

        >>> a = IntRangeSet('0:5,6:10')
        >>> a.add('5','100:200')
        >>> print(a)
        IntRangeSet('0:10,100:200')
        """

        # !!consider special casing the add of a single int. Anything else?
        for start, stop in IntRangeSet._static_ranges(*ranges_inputs):
            self._internal_add(start, stop - start)

    def __iadd__(self, *ranges_inputs):
        """See :meth:`IntRangeSet.add`"""
        self.add(*ranges_inputs)
        return self

    __ior__ = __iadd__
    update = add

    def __imul__(self, n):
        """
        Put produces n copies of a unioned together in a, where a is an IntRangeSet.
        Because a is a set, the result will either be an empty IntRangeSet (n is 0 or less) or the original a.

        * ``a *= n``

        :Example:

        >>> a = IntRangeSet('0:5,6:11')
        >>> a *= 5
        >>> print(a) # should be unchanged
        IntRangeSet('0:5,6:11')
        >>> a *= 0
        >>> print(a) # should be empty
        IntRangeSet('')
        """
        if n <= 0:
            self.clear()
        return self

    def copy(self):
        """
        Create a deep copy of a IntRangeSet.
        """
        return IntRangeSet(self)

    def ranges(self):
        """
        Iterate, in order, the ranges of a IntRangeSet as (start,stop) tuples.

        :Example:

        >>> for start,stop in IntRangeSet('0:10,100:200').ranges():
        ...       print("start is {0}, stop is {1}".format(start,stop))
        start is 0, stop is 10
        start is 100, stop is 200

        """
        for item in self._start_items:
            stop = item + self._start_to_length[item]
            yield item, stop

    def __iter__(self):
        """
        Iterate, in order from smallest to largest, the integer elements of the IntRangeSet

        :Example:

        >>> for i in IntRangeSet('1:4,10'):
        ...    print(i)
        1
        2
        3
        10

        """
        for first, stop in self.ranges():
            for i in range(first, stop):
                yield i

    def clear(self):
        """
        Remove all ranges from this IntRangeSet.

        >>> a = IntRangeSet('0:10,12')
        >>> a.clear()
        >>> print(a)
        IntRangeSet('')

        """
        del self._start_items[:]
        self._start_to_length.clear()

    def __len__(self):
        """
        The number of integer elements in the IntRangeSet

        >>> print(len(IntRangeSet('0:10,12')))
        11

        Note: This is computed in time linear in the number of ranges, rather than integer elements.

        """
        return sum(self._start_to_length.values())

    @property
    def ranges_len(self):
        """
        The number of contiguous ranges in the IntRangeSet

        >>> print(IntRangeSet('0:10,12').ranges_len)
        2

        """
        return len(self._start_items)

    def ranges_getitem(self, index):
        """
        Returns the index-th range in the collection of ranges

        >>> print(IntRangeSet('-30:-20,0:10,12').ranges_getitem(1))
        (0, 10)

        """
        start = self._start_items[index]
        stop = start + self._start_to_length[start]
        return (start, stop)

    def ranges_index(self, element):
        """
        Returns the ranges index of range containing the element

        >>> int_range_set = IntRangeSet('-30:-20,0:10,12')
        >>> index = int_range_set.ranges_index(5)
        >>> print(index)
        1
        >>> int_range_set.ranges_getitem(index)
        (0, 10)
        """
        index = bisect_left(self._start_items, element)
        if index == len(self._start_items) or self._start_items[index] != element:
            index -= 1
        start, stop = self.ranges_getitem(index)
        if element < start or stop <= element:
            raise ValueError("element Not Found")
        return index

    def sum(self):
        """
        The sum of the integer elements in the IntRangeSet

        >>> print(IntRangeSet('0:10,12').sum())
        57

        Note: This is more efficient than ``sum(IntRangeSet('0:10,12'))`` because is computed
        in time linear in the number of ranges, rather than integer elements.
        """
        result = 0
        for start in self._start_items:
            length = self._start_to_length[start]
            result += (start + start + length - 1) * length // 2
        return result

    def __eq__(self, other):  # !!  'others' to ranges_input
        """
        True exactly when the IntRangeSet on the left is *set equivalent* to the ranges input on the right.

        * ``a == b``


        >>> print(IntRangeSet('0:10,12') == IntRangeSet('0:10,12'))
        True
        >>> print(IntRangeSet('0:10,12') == IntRangeSet('0:10'))
        False
        >>> print(IntRangeSet('0:10,12') == IntRangeSet('12,0:5,5:10'))
        True
        >>> print(IntRangeSet('0:10,12') == '0:10,12') # The right-hand can be any ranges input
        True
        """
        self, other = IntRangeSet._make_args_range_set(self, other)
        if other is None or len(self._start_items) != len(other._start_items):
            return False
        for i, start in enumerate(self._start_items):
            if (
                start != other._start_items[i]
                or self._start_to_length[start] != other._start_to_length[start]
            ):
                return False
        return True

    def __ne__(self, other):
        """
        *False* exactly when the IntRangeSet on the left is *set equivalent* to the ranges input on the right.

        These are the same.

        * ``a != b``
        * ``not a == b``
        """
        # Don't need to call _make_args_range_set because __eq__ will call it
        return not self == other

    # Same: a >= b, a.issuperset(b,...), b in a
    # Returns True iff item is within the ranges of this IntRangeSet.
    def __contains__(self, *ranges_inputs):
        """
        True exactly when all the ranges input is a subset of the IntRangeSet.

        These are the same:

        * ``b in a``
        * ``a.issuperset(b)``
        * ``a >= b``


        :Example:

        >>> print(3 in IntRangeSet('0:5,6:11'))
        True
        >>> print(IntRangeSet('4:7') in IntRangeSet('0:5,6:11'))
        False
        >>> '6:9' in IntRangeSet('0:5,6:11') # The left-hand of 'in' can be any ranges input
        True
        >>> print(IntRangeSet('0:5,6:11') >= '6:9') # The right-hand of can be any ranges input
        True

        The 'issuperset' method also supports unioning multiple ranges inputs.

        :Example:

        >>> print(IntRangeSet('0:5,6:11').issuperset(4,7,8))
        True
        >>> print(IntRangeSet('0:5,6:11').issuperset(4,7,8,100))
        False

        Note: By definition, any set is a superset of itself.
        """
        for start_in, stop_in in IntRangeSet._static_ranges(*ranges_inputs):
            (
                start_self,
                length_self,
                index,
                contains,
            ) = self._best_start_length_index_contains(start_in)
            if not contains or stop_in > start_self + length_self:
                return False
        return True

    def __ge__(self, other):
        """See :meth:`IntRangeSet.__contains__`"""
        return other in self

    issuperset = __contains__

    @property
    def isempty(self):
        """
        True exactly when the IntRangeSet is empty.

        >>> print(IntRangeSet().isempty)
        True
        >>> print(IntRangeSet(4).isempty)
        False
        """
        return len(self._start_items) == 0

    def __str__(self):
        """
        Use the standard str(a) function to create a string representation of a, an IntRangeSet.

        >>> print("Hello " + str(IntRangeSet(2,3,4,10)))
        Hello IntRangeSet('2:5,10')
        """
        return repr(self)

    def __repr__(self):
        """
        Use the standard repr(a) function to create a string representation of a, an IntRangeSet.

        >>> print("Hello " + repr(IntRangeSet(2,3,4,10)))
        Hello IntRangeSet('2:5,10')
        """
        return "IntRangeSet('{0}')".format(self._repr_internal(":", ","))

    def _repr_internal(self, seperator1, separator2):
        if self.isempty:
            return ""

        fp = io.StringIO() if sys.version_info >= (3, 0) else io.BytesIO()

        for index, (start, stop) in enumerate(self.ranges()):
            if index > 0:
                fp.write(separator2)

            if start == stop - 1:
                fp.write(str(start))
            else:
                fp.write("{0}{1}{2}".format(start, seperator1, stop))
        return fp.getvalue()

    @staticmethod
    def _test():
        assert IntRangeSet("100:110,1000").index("109,100:104") == "0:4,9"
        assert IntRangeSet("-3") == "-3"
        int_range_set = IntRangeSet()
        int_range_set.add(0)
        assert "IntRangeSet('0')" == str(int_range_set)
        int_range_set.add(1)
        assert "IntRangeSet('0:2')" == str(int_range_set)
        int_range_set.add(4)
        assert "IntRangeSet('0:2,4')" == str(int_range_set)
        int_range_set.add(5)
        assert "IntRangeSet('0:2,4:6')" == str(int_range_set)
        int_range_set.add(7)
        assert "IntRangeSet('0:2,4:6,7')" == str(int_range_set)
        int_range_set.add(2)
        assert "IntRangeSet('0:3,4:6,7')" == str(int_range_set)
        int_range_set.add(3)
        assert "IntRangeSet('0:6,7')" == str(int_range_set)
        int_range_set.add(6)
        assert "IntRangeSet('0:8')" == str(int_range_set)
        int_range_set.add(-10)
        assert "IntRangeSet('-10,0:8')" == str(int_range_set)
        int_range_set.add(-5)
        assert "IntRangeSet('-10,-5,0:8')" == str(int_range_set)

        assert IntRangeSet("-10:-4") == "-10:-4"
        assert IntRangeSet("-10:-4,-3") == "-10:-4,-3"
        assert IntRangeSet("-10:-4,-3,-2:2") == "-10:-4,-3:2"
        assert IntRangeSet("-10:-4,-3,-2:2,1:6") == "-10:-4,-3:6"
        assert IntRangeSet("-10:-4,-3,-2:2,1:6,7:13") == "-10:-4,-3:6,7:13"
        assert IntRangeSet("-10:-4,-3,-2:2,1:6,7:13,13:16") == "-10:-4,-3:6,7:16"
        assert IntRangeSet("-10:-4,-3,-2:2,1:6,7:13,13:16,14:17") == "-10:-4,-3:6,7:17"
        assert (
            IntRangeSet("-10:-4,-3,-2:2,1:6,7:13,13:16,14:17,20:26")
            == "-10:-4,-3:6,7:17,20:26"
        )
        assert (
            IntRangeSet("-10:-4,-3,-2:2,1:6,7:13,13:16,14:17,20:26,22:24")
            == "-10:-4,-3:6,7:17,20:26"
        )

        s = "-10:-4,-3,-2:2,1:6,7:13,13:16,14:17,20:26,22:24"
        int_range_set = IntRangeSet(s)
        assert int_range_set == "-10:-4,-3:6,7:17,20:26"

        s = "1:6,0,4:11,-10:-4,-12:-2,15:21,12:22,-13"
        int_range_set = IntRangeSet(s)
        assert int_range_set == "-13:-2,0:11,12:22"

        assert len(int_range_set) == 32

        int_range_set1 = IntRangeSet("-10:-4")
        int_range_set2 = int_range_set1.copy()
        assert int_range_set1 is not int_range_set2
        assert int_range_set1 == int_range_set2
        int_range_set2.add(7)
        assert int_range_set1 != int_range_set2

        assert str(IntRangeSet(7)) == "IntRangeSet('7')"
        assert str(IntRangeSet((7, 8))) == "IntRangeSet('7')"
        assert str(IntRangeSet((7, 11))) == "IntRangeSet('7:11')"
        assert str(IntRangeSet(list(range(7, 11)))) == "IntRangeSet('7:11')"
        assert str(IntRangeSet(np.s_[7:11])) == "IntRangeSet('7:11')"
        assert str(IntRangeSet(np.s_[7:11:2])) == "IntRangeSet('7,9')"
        assert str(IntRangeSet(list(range(7, 11, 2)))) == "IntRangeSet('7,9')"
        assert str(IntRangeSet(None)) == "IntRangeSet('')"
        assert str(IntRangeSet()) == "IntRangeSet('')"
        assert [e for e in IntRangeSet("-10:-4,-3")] == [-10, -9, -8, -7, -6, -5, -3]
        int_range_set3 = IntRangeSet(7, 10)
        int_range_set3.clear()
        assert str(int_range_set3) == "IntRangeSet('')"
        assert len(IntRangeSet("-10:-4,-3")) == 7

        int_range_set4 = IntRangeSet("-10:-4,-3")
        int_range_set4.add(-10, -7)
        assert int_range_set4 == "-10:-4,-3"
        int_range_set4.add(-10, -4)
        assert int_range_set4 == "-10:-2"

        int_range_set5 = IntRangeSet("-10:-4,-3")
        assert -11 not in int_range_set5
        assert -10 in int_range_set5
        assert -5 in int_range_set5
        assert -4 not in int_range_set5
        assert -3 in int_range_set5
        assert -2 not in int_range_set5
        assert 19999 not in int_range_set5
        assert "-11" not in int_range_set5
        assert "-10" in int_range_set5
        assert "-10:-4" in int_range_set5
        assert "-10:-3" not in int_range_set5
        assert "-10:-4,-3" in int_range_set5
        assert "-10:-4,-3,100" not in int_range_set5
        assert IntRangeSet("-11") not in int_range_set5
        assert IntRangeSet("-10") in int_range_set5
        assert IntRangeSet("-10:-4") in int_range_set5
        assert IntRangeSet("-10:-3") not in int_range_set5
        assert IntRangeSet("-10:-4,-3") in int_range_set5
        assert IntRangeSet("-10:-4,-3,100") not in int_range_set5
        assert [-11] not in int_range_set5
        assert [-10] in int_range_set5
        assert list(range(-10, -6)) in int_range_set5
        assert list(range(-10, -3)) not in int_range_set5
        assert [-10, -9, -8, -7, -3] in int_range_set5
        assert [-10, -9, -8, -7, -3, -100] not in int_range_set5

        assert IntRangeSet("-11,-10,-9") == IntRangeSet("-11:-8")
        a = IntRangeSet("-11,-10")
        b = IntRangeSet("-11:-8")
        assert a != b
        assert IntRangeSet("-11,1") != IntRangeSet("-11:1")

        assert IntRangeSet("1:4") | IntRangeSet("2,4,6") == IntRangeSet("1:5,6")
        assert IntRangeSet("1:4").union("2,4,6", "1,6,-100") == IntRangeSet(
            "-100,1:5,6"
        )

        assert IntRangeSet() & IntRangeSet() == IntRangeSet()
        assert IntRangeSet().intersection("", "") == IntRangeSet()
        assert IntRangeSet() & IntRangeSet() & IntRangeSet() == IntRangeSet()

        assert IntRangeSet("1") & IntRangeSet() == IntRangeSet()
        assert IntRangeSet() & IntRangeSet("1") == IntRangeSet()
        assert IntRangeSet("1:6") & IntRangeSet() == IntRangeSet()
        assert IntRangeSet() & IntRangeSet("1:6") == IntRangeSet()
        assert IntRangeSet("1:6,7") & IntRangeSet() == IntRangeSet()
        assert IntRangeSet() & IntRangeSet("1:6,7") == IntRangeSet()

        assert IntRangeSet("1") & IntRangeSet("1") == IntRangeSet("1")
        assert IntRangeSet("1:6") & IntRangeSet("1") == IntRangeSet("1")
        assert IntRangeSet("1") & IntRangeSet("1:6") == IntRangeSet("1")
        assert IntRangeSet("1:6,7") & IntRangeSet("1") == IntRangeSet("1")

        assert IntRangeSet("2") & IntRangeSet("1:6,7") == IntRangeSet("2")
        assert IntRangeSet("1:6") & IntRangeSet("2") == IntRangeSet("2")
        assert IntRangeSet("2") & IntRangeSet("1:6") == IntRangeSet("2")
        assert IntRangeSet("1:6,7") & IntRangeSet("2") == IntRangeSet("2")

        assert IntRangeSet("-2") & IntRangeSet("1:6,7") == IntRangeSet()
        assert IntRangeSet("1:6") & IntRangeSet("-2") == IntRangeSet()
        assert IntRangeSet("-2") & IntRangeSet("1:6") == IntRangeSet()
        assert IntRangeSet("1:6,7") & IntRangeSet("-2") == IntRangeSet()

        assert IntRangeSet("22") & IntRangeSet("1:6,7") == IntRangeSet()
        assert IntRangeSet("1:6") & IntRangeSet("22") == IntRangeSet()
        assert IntRangeSet("22") & IntRangeSet("1:6") == IntRangeSet()
        assert IntRangeSet("1:6,7") & IntRangeSet("22") == IntRangeSet()

        assert IntRangeSet("-2,1:4,20,25:100,101") & IntRangeSet(
            "1:101"
        ) == IntRangeSet("1:4,20,25:100")
        assert IntRangeSet("2:4,90:110") & IntRangeSet("1:100") == IntRangeSet(
            "2:4,90:100"
        )
        assert IntRangeSet("1:100") & IntRangeSet(
            "-2,1:4,20,25:100,100"
        ) == IntRangeSet("1:4,20,25:100")
        assert IntRangeSet("1:100") & IntRangeSet("2:4,90:110") == IntRangeSet(
            "2:4,90:100"
        )

        assert IntRangeSet("0:67") - "30:101" == IntRangeSet("0:30")
        assert IntRangeSet("0:30,51:67").difference("40:101") == IntRangeSet("0:30")
        assert IntRangeSet("0:67").difference("30:50", "40:100") == IntRangeSet("0:30")
        assert IntRangeSet("0:67,99,200") - "30:100,300" == "0:30,200"
        assert IntRangeSet("30:100") - "0:67" == "67:100"
        assert IntRangeSet("30:100,300") - IntRangeSet("0:67,100,200") == IntRangeSet(
            "67:100,300"
        )

        assert IntRangeSet("30:100,300") ^ IntRangeSet("0:67,99,200") == IntRangeSet(
            "0:30,67:99,200,300"
        )
        assert IntRangeSet([1, 2]).symmetric_difference([2, 3]) == IntRangeSet([1, 3])

        assert IntRangeSet("10:15")[0:1] == IntRangeSet("10")
        assert IntRangeSet("10:15,55:60")[0:5] == IntRangeSet("10:15")
        assert IntRangeSet("10:15,55:60")[1:5] == IntRangeSet("11:15")
        assert IntRangeSet("10:15,55:60")[8:9] == IntRangeSet("58")
        assert IntRangeSet("10:15,55:60")[-2:] == IntRangeSet("58:60")
        assert IntRangeSet("10:15,55:60")[0:5:2] == IntRangeSet("10,12,14")

        try:
            IntRangeSet("10:15,55:60")[100]
        except KeyError:
            pass
        try:
            IntRangeSet("10:15,55:60")[-100]
        except KeyError:
            pass

        assert IntRangeSet("10:15,55:60").max() == 59
        assert IntRangeSet("10:15,55:60").min() == 10

        assert IntRangeSet("10:15,55:60") + IntRangeSet("58:61,100") == IntRangeSet(
            "10:15,55:61,100"
        )
        assert IntRangeSet("10:15,55:60") + "58:61,100" == IntRangeSet(
            "10:15,55:61,100"
        )

        mult_test0 = IntRangeSet("10:15,55:60")
        mult_test1 = mult_test0 * 3
        assert mult_test0 == mult_test1
        assert mult_test0 is not mult_test1
        assert (mult_test0 * 0).isempty
        assert (mult_test0 * -1000).isempty

        assert IntRangeSet("10:15,55:60").index(10) == 0
        assert IntRangeSet("10:15,55:60").index(11) == 1
        assert IntRangeSet("10:15,55:60").index(55) == 5
        assert IntRangeSet("10:15,55:60").index(56) == 6
        try:
            IntRangeSet("10:15,55:60").index(9)
        except IndexError:
            pass
        try:
            IntRangeSet("10:15,55:60").index(15)
        except IndexError:
            pass

        assert (
            IntRangeSet("10:15,55:60").index("57,56:57") == "6:8"
        )  # returns the index of the start of the contiguous place where 57,"56-56" occurs
        assert IntRangeSet("10:15,55:60").index([10, 55]) == "0,5"
        try:
            IntRangeSet("10:15,55:60").index("100:111")
        except IndexError:
            pass
        try:
            IntRangeSet("10:15,55:60").index("14:17")
        except IndexError:
            pass

        try:
            IntRangeSet(3.34)
        except Exception:
            pass
        try:
            3.34 in IntRangeSet()
        except Exception:
            pass

        assert IntRangeSet("10:15,55:60").count(10) == 1
        assert IntRangeSet("10:15,55:60").count(100) == 0

        assert IntRangeSet("10:15,55:60").isdisjoint(100)
        assert not IntRangeSet("10:15,55:60").isdisjoint("57:100")

        assert IntRangeSet("10:15,55:60") <= "10:15,55:60"
        assert IntRangeSet("10:15,55:60").issubset("10:15,55:60")
        assert not IntRangeSet("10:15,55:60") < "10:15,55:60"
        assert IntRangeSet("10:15,55:60") < "9:15,55:60"

        assert IntRangeSet("10:15,55:60") >= "10:15,55:60"
        assert IntRangeSet("10:15,55:60").issuperset("10:15,55:60")
        assert not IntRangeSet("10:15,55:60") > "10:15,55:60"
        assert IntRangeSet("9:15,55:60") > "10:15,55:60"

        update0 = IntRangeSet("9:15,55:60")
        update0.update("10,30,100", "30")
        assert update0 == IntRangeSet("9:15,30,55:60,100")

        assert IntRangeSet() is not None

        update0 = IntRangeSet()
        update0 |= []
        assert update0 == IntRangeSet()

        update0 = IntRangeSet("9:15,55:60")
        update0.intersection_update("10,100", "0:101")
        assert update0 == IntRangeSet("10")

        update0 = IntRangeSet("9:15,55:60")
        update0 |= IntRangeSet("10,30,100") | "30"
        assert update0 == IntRangeSet("9:15,30,55:60,100")

        update0 = IntRangeSet("9:15,55:60")
        update0 += IntRangeSet("10,30,100") + "30"
        assert update0 == IntRangeSet("9:15,30,55:60,100")

        update0 = IntRangeSet("9:15,55:60")
        update0 &= IntRangeSet("10,100") & "0:101"
        assert update0 == IntRangeSet("10")

        update0 = IntRangeSet("9:15,55:60")
        update0 -= IntRangeSet("10,100") - "30"
        assert update0 == IntRangeSet("9,11:15,55:60")

        update0 = IntRangeSet("9:15,55:60")
        update0.difference_update("10,100", "30")
        assert update0 == IntRangeSet("9,11:15,55:60")

        update0 = IntRangeSet("9:15,55:60")
        update0.symmetric_difference_update("10,100,30")
        assert update0 == IntRangeSet("9,11:15,30,55:60,100")

        update0 = IntRangeSet("9:15,55:60")
        update0 ^= IntRangeSet("10,100,30")
        assert update0 == IntRangeSet("9,11:15,30,55:60,100")

        remove0 = IntRangeSet("9:15,55:60")
        remove0.remove(9)
        assert remove0 == IntRangeSet("10:15,55:60")

        remove0 = IntRangeSet("9:15,55:60")
        remove0.remove(10, 13)
        assert remove0 == IntRangeSet("9,11:13,14,55:60")

        remove0 = IntRangeSet("9:15,55:60")
        try:
            remove0.remove("100")
        except KeyError:
            pass

        remove0 = IntRangeSet("9:15,55:60")
        try:
            remove0.remove(IntRangeSet("100"), 101)
        except Exception:
            pass

        discard0 = IntRangeSet("9:15,55:60")
        discard0.discard(9)
        assert discard0 == IntRangeSet("10:15,55:60")

        discard0 = IntRangeSet("9:15,55:60")
        discard0.discard(10, 13)
        assert discard0 == IntRangeSet("9,11:13,14,55:60")

        discard0 = IntRangeSet("9:15,55:60")
        discard0.discard("100")
        assert discard0 == "9:15,55:60"

        discard0 = IntRangeSet("9:15,55:60")
        discard0.discard(IntRangeSet("100"), 101)
        assert discard0 == "9:15,55:60"

        pop0 = IntRangeSet("9:15,55:60")
        assert pop0.pop() == 59
        assert pop0 == "9:15,55:59"

        pop0 = IntRangeSet([1, 3])
        assert pop0.pop() == 3
        assert pop0.pop() == 1
        try:
            pop0.pop()
        except KeyError:
            pass

        delitem0 = IntRangeSet("10:15,55:60")
        del delitem0[0:5]
        delitem0 == "55:60"

        delitem0 = IntRangeSet("10:15,55:60")
        del delitem0[1:5]
        delitem0 == "10,55:60"

        delitem0 = IntRangeSet("10:15,55:60")
        del delitem0[9:10]
        delitem0 == "10:15,55:59"

        delitem0 = IntRangeSet("10:15,55:60")
        del delitem0[-2:]
        delitem0 == "10:15,55:58"

        delitem0 = IntRangeSet("10:15,55:60")
        del delitem0[0:5:2]
        delitem0 == "11,33,55:60"

        delitem0 = IntRangeSet("10:15,55:60")
        del delitem0[-2]
        delitem0 == "10:15,55:58,59"

        delitem0 = IntRangeSet("10:15,55:60")
        del delitem0[3]
        delitem0 == "10:13,14,55:60"

        delitem0 = IntRangeSet("10:15,55:60")
        try:
            del delitem0[100]
        except KeyError:
            pass
        try:
            del delitem0[-100]
        except KeyError:
            pass

        assert list(reversed(IntRangeSet("10:15,55:60"))) == list(
            reversed(list(IntRangeSet("10:15,55:60")))
        )

        IntRangeSet("10:15,55:60").sum() == sum(IntRangeSet("10:15,55:60"))

        add0 = IntRangeSet("1,12:15,55:61,71,102")
        add0.add("12:101")
        assert add0 == "1,12:101,102"

        assert IntRangeSet("1,12:15,55:61,71,102") - "5:72" == "1,102"
        assert IntRangeSet("1,12:15,55:61,71,102") - "12:66" == "1,71,102"
        assert IntRangeSet("1,12:15,55:61,71,102") - "13:57" == "1,12,57:61,71,102"

        a = IntRangeSet("100:200,1000")
        del a["2:11"]
        assert a == "100:102,111:200,1000"

        assert IntRangeSet("0:6,6:10") - "3:100" == "0:3"

        a = IntRangeSet("100:200")
        a *= 3
        assert a == IntRangeSet("100:200")
        a *= 0
        assert a.isempty
        a = IntRangeSet("100:200")
        a *= -1
        assert a.isempty

        assert IntRangeSet("").ranges_len == 0
        assert IntRangeSet("3").ranges_len == 1
        assert IntRangeSet("3,6").ranges_len == 2
        assert IntRangeSet("3,4,5,6").ranges_len == 1

        assert IntRangeSet("100:200")["10:20,30"] == IntRangeSet("110:120,130")

        try:
            IntRangeSet("100..200")  # not well formed
        except Exception:
            pass

        assert IntRangeSet("1,12:15,55:61,71,102").ranges_getitem(1) == (12, 15)
        assert IntRangeSet("1,12:15,55:61,71,102").ranges_getitem(-1) == (102, 103)
        try:
            IntRangeSet("1,12:15,55:61,71,102").ranges_getitem(5)
        except Exception:
            pass
        assert IntRangeSet("1,12:15,55:61,71,102").ranges_index(13) == 1
        try:
            assert IntRangeSet("1,12:15,55:61,71,102").ranges_index(13) == 1
        except Exception:
            pass
        try:
            IntRangeSet("1,12:15,55:61,71,102").ranges_index(2)
        except Exception:
            pass
        try:
            IntRangeSet("1,12:15,55:61,71,102").ranges_index(103)
        except Exception:
            pass
        assert IntRangeSet("1,12:15,55:61,71,102").ranges_index(1) == 0
        assert IntRangeSet("1,12:15,55:61,71,102").ranges_index(102) == 4

    # s[i] ith item of s, origin 0 (3)
    # s[i:j] slice of s from i to j (3)(4)
    # s[i:j:k] slice of s from i to j with step k (3)(5)
    def __getitem__(self, key):
        """
        ``a[i]`` returns the ith integer in sorted order (origin 0) from a, an IntRangeSet

        >>> print(IntRangeSet('100:200,1000')[0])
        100
        >>> print(IntRangeSet('100:200,1000')[10])
        110

        If i is negative, the indexing goes from the end

        >>> print(IntRangeSet('100:200,1000')[-1])
        1000

        Python's standard slice notation may be used and returns IntRangeSets.
        (Remember that the Stop number in slice notation is exclusive.)

        >>> print(IntRangeSet('100:200,1000')[0:10]) # Integers 0 (inclusive) to 10 (exclusive)
        IntRangeSet('100:110')

        >>> print(IntRangeSet('100:200,1000')[0:10:2]) # Integers 0 (inclusive) to 10 (exclusive) with step 2
        IntRangeSet('100,102,104,106,108')

        >>> print(IntRangeSet('100:200,1000')[-3:]) # The last three integers in the IntRangeSet.
        IntRangeSet('198:200,1000')

        An IntRangeSet can also be accessed with any ranges input.

        >>> IntRangeSet('100:200,1000')['0:10,20']
        IntRangeSet('100:110,120')
        """
        if isinstance(key, numbers.Integral):
            if key >= 0:
                for start in self._start_items:
                    length = self._start_to_length[start]
                    if key < length:
                        return start + key
                    key -= length
                raise KeyError()
            else:
                assert key < 0
                key = -key - 1
                for start_index in range(len(self._start_items)):
                    start = self._start_items[-1 - start_index]
                    length = self._start_to_length[start]
                    if key < length:
                        return start + length - 1 - key
                    key -= length
                raise KeyError()
        elif isinstance(key, slice):
            lenx = len(self)
            start_index, stop_index, step_index = key.start, key.stop, key.step
            start_index = start_index or 0
            stop_index = stop_index or lenx
            step_index = step_index or 1

            if step_index == 1:
                return self & (self[start_index], self[stop_index - 1] + 1)
            else:
                return IntRangeSet(self[index] for index in range(*key.indices(lenx)))
        else:
            start_and_stop_generator = (
                self._two_index(start_index, stop_index)
                for start_index, stop_index in IntRangeSet._static_ranges(key)
            )
            return self.intersection(start_and_stop_generator)

    # max(s) largest item of s
    def max(self):
        """
        The largest integer element in the IntRangeSet

        >>> print(IntRangeSet('0:10,12').max())
        12

        Note: This is more efficient than max(IntRangeSet('0:10,12')) because is computed
        in constant time rather than in time linear to the number of integer elements.
        """
        start = self._start_items[-1]
        return start + self._start_to_length[start] - 1

    # min(s) smallest item of s
    def min(self):
        """
        The smallest integer element in the IntRangeSet

        :Example:

        >>> print(IntRangeSet('0:10,12').min())
        0

        Note: This is more efficient than ``min(IntRangeSet('0:10,12'))`` because is computed
        in constant time rather than in time linear to the number of integer elements.
        """
        return self._start_items[0]

    def _make_args_range_set(*args):
        for arg in args:
            if arg is None:
                yield None
            elif isinstance(arg, IntRangeSet):
                yield arg
            else:
                yield IntRangeSet(arg)

    def union(*ranges_inputs):
        """
        Return the union of a IntRangeSet with zero or more ranges inputs. The original IntRangeSet is not changed.

        These are the same:

        * ``a | b``
        * ``a + b``
        * ``a.union(b)``

        :Example:

        >>> print(IntRangeSet('0:5,6:10') | 5)
        IntRangeSet('0:10')

        The 'union' method also support unioning multiple ranges inputs,

        :Example:

        >>> print(IntRangeSet('0:5,6:10').union(5,'100:200'))
        IntRangeSet('0:10,100:200')
        """
        result = IntRangeSet()
        result.add(*ranges_inputs)
        return result

    __or__ = union
    __add__ = union

    # s * n, n shallow copies of s concatenated (2)
    def __mul__(self, n):
        """
        ``a * n``, produces n shallow copies of a unioned, where a is an IntRangeSet.
        Because a is a set, the result will either be an empty IntRangeSet (n is 0 or less) or a copy of
        the original IntRangeSet.

        * ``a * n``

        """
        if n <= 0:
            return IntRangeSet()
        else:
            return IntRangeSet(self)

    def index(self, other):
        """
        If x is an integer, returns the index of x in a, where a is an IntRangeSet.
        If x is an ranges input, returns an IntRangeSet of index of every integer in x.
        Raises an IndexError is x not in a.

        ``* a.index(x)``


        >>> print(IntRangeSet('100:110,1000').index(109))
        9
        >>> print(IntRangeSet('100:110,1000').index('109,100:104'))
        IntRangeSet('0:4,9')
        """
        if isinstance(other, numbers.Integral):
            return self._index_element(other)
        else:
            # If start and stop are adjacent, only call _index_element once
            start_and_stop_index_generator = (
                (self._index_element(start), self._index_element(stop - 1) + 1)
                for start, stop in IntRangeSet._static_ranges(other)
            )
            return IntRangeSet(start_and_stop_index_generator)

    def _index_element(self, element):
        index = bisect_left(self._start_items, element)

        # Find the position_in_its_range of this element
        if (
            index != len(self._start_items) and self._start_items[index] == element
        ):  # element is start value
            position_in_its_range = 0
        elif index == 0:  # item is before any of the ranges
            raise IndexError()
        else:
            index -= 1
            start = self._start_items[index]
            stop = start + self._start_to_length[start]
            if element >= stop:  # we already know it's greater than start
                raise IndexError()
            else:
                position_in_its_range = element - start

        # Sum up the length of all preceding ranges
        preceeding_starts = self._start_items[0:index]
        preceeding_lengths = (
            self._start_to_length[start] for start in preceeding_starts
        )
        result = position_in_its_range + sum(preceeding_lengths)
        return result

    # s.count(x) total number of occurrences of x in s
    def count(self, ranges):
        """
        The number of times that the elements of ranges appears in the IntRangeSet. Because IntRangeSet is
        a set, the number will be either 0 or 1.

        >>> print(IntRangeSet('100:110,1000').count('105:107,1000'))
        1
        """
        if ranges in self:
            return 1
        else:
            return 0

    # Return True if the set has no elements in common with other. Sets are disjoint if and only if their intersection is the empty set.
    def isdisjoint(self, ranges):
        """
        True exactly when the two sets have no integer elements in common.

        :Example:

        >>> print(IntRangeSet('100:110,1000').isdisjoint('900:2000'))
        False
        >>> print(IntRangeSet('100:110,1000').isdisjoint('1900:2000'))
        True
        """
        isempty_generator = (
            IntRangeSet(tuple)._binary_intersection(self).isempty
            for tuple in IntRangeSet._static_ranges(ranges)
        )
        return all(isempty_generator)

    def __le__(self, ranges):
        """
        True exactly when the IntRangeSet is a subset of the ranges.

        These are the same:

        * ``a <= b``
        * ``a.issubset(b)``

        :Example:

        >>> print(IntRangeSet('0:5,6:11') <= '-1:101') # The right-hand can be any ranges input
        True

        Note: By definition, any set is a subset of itself.
        """
        self, ranges = IntRangeSet._make_args_range_set(self, ranges)
        return self in ranges

    issubset = __le__

    # set < other
    # Test whether the set is a proper subset of other, that is, set <= other and set != other.
    def __lt__(self, ranges):
        """
        True exactly when the IntRangeSet is a proper subset of the ranges.

        * ``a < b``

        :Example:

        >>> print(IntRangeSet('0:5,6:11') < '-1:101') # The right-hand can be any ranges input
        True

        Note: By definition, no set is a proper subset of itself.
        """
        self, ranges = IntRangeSet._make_args_range_set(self, ranges)
        return self != ranges and self in ranges

    # set > other
    # Test whether the set is a proper superset of other, that is, set >= other and set != other.
    def __gt__(self, other):
        """
        True exactly when the IntRangeSet is a proper superset of the ranges.

        * ``a > b``

        :Example:

        >>> print(IntRangeSet('0:5,6:11') > '7:10') # The right-hand can be any ranges input
        True

        Note: By definition, no set is a proper superset of itself.
        """
        self, other = IntRangeSet._make_args_range_set(self, other)
        return self != other and other in self

    def intersection(*ranges_inputs):
        """
        Return the intersection of a IntRangeSet and zero or more ranges inputs. The original IntRangeSet is not changed.

        These are the same:

        * ``a & b``
        * ``a.intersection(b)``

        :Example:

        >>> print(IntRangeSet('0:5,6:11') & '3:8')
        IntRangeSet('3:5,6:8')

        The 'intersection' method also support intersecting multiple ranges inputs,

        :Example:

        >>> print(IntRangeSet('0:5,6:11').intersection('3:8','4:7'))
        IntRangeSet('4,6')
        """
        ranges_inputs = IntRangeSet._make_args_range_set(
            *ranges_inputs
        )  # generator to made every ranges a IntRangeSet
        ranges_inputs = sorted(
            ranges_inputs, key=lambda int_range_set: len(int_range_set._start_items)
        )  # sort so that IntRangeSet with smaller range_count is first
        result = ranges_inputs[0]  # LATER what if no args, empty? The universe?
        for ranges in ranges_inputs[1:]:
            result = result._binary_intersection(ranges)
        return result

    __and__ = intersection

    def _binary_intersection(self, other):
        result = IntRangeSet()

        if self.isempty:
            return result

        index = 0
        start0 = self._start_items[index]
        length0 = self._start_to_length[start0]
        stop0 = start0 + length0
        while True:
            start1, length1, index1, contains = other._best_start_length_index_contains(
                start0
            )
            stop1 = start1 + length1
            if contains:
                if (
                    stop0 <= stop1
                ):  # All of range0 fits inside some range1, so add it the intersection and next look at the next range0
                    result._internal_add(start0, length0)
                    index += 1
                    if index >= len(self._start_items):
                        break  # leave the while loop
                    start0 = self._start_items[index]
                    length0 = self._start_to_length[start0]
                    stop0 = start0 + length0
                else:  # Only part of range0 fits inside some range0, so add that part and then next look at the rest of range0
                    result._internal_add(start0, stop1 - start0)
                    start0 = stop1
                    length0 = stop0 - start0
            else:  # start0 is not contained in any range1, so swap self and other and then next look at the next range0
                temp = other
                other = self
                self = temp
                index = index1 + 1
                if index >= len(self._start_items):
                    break  # leave the while loop
                start0 = self._start_items[index]
                length0 = self._start_to_length[start0]
                stop0 = start0 + length0
        return result

    # Same a-b, a.difference(b,...)
    # difference(other, ...)set - other - ...
    # Return a new set with elements in the set that are not in the others.
    # Changed in version 2.6: Accepts multiple input iterables.
    def __sub__(
        self, *ranges_inputs
    ):  # !!could be made faster by being more direct instead of using complements
        """
        Return the set difference of a IntRangeSet with zero or more ranges inputs. The original IntRangeSet is not changed.

        These are the same:

        * ``a - b``
        * ``a.difference(b)``

        :Example:

        >>> print(IntRangeSet('0:5,6:11') - 1)
        IntRangeSet('0,2:5,6:11')
        >>> print(IntRangeSet('0:5,6:11') - '3:100')
        IntRangeSet('0:3')

        The 'difference' method also supports subtracting multiple input ranges

        :Example:

        >>> print(IntRangeSet('0:5,6:11').difference('3:100',1))
        IntRangeSet('0,2')
        """
        result = self.copy()
        result.difference_update(*ranges_inputs)
        return result

    difference = __sub__

    # same a^b, a.symmetric_difference(b)
    # symmetric_difference(other)set ^ other
    # Return a new set with elements in either the set or other but not both.
    def __xor__(self, ranges):
        """
        Returns a new IntRangeSet set with elements in either the input IntRangeSet or the input range but not both.

        These are the same:

        * ``a ^ b``
        * ``a.symmetric_difference(b)``

        :Example:

        >>> print(IntRangeSet('0:5,6:11') ^ '3:9')
        IntRangeSet('0:3,5,9:11')
        """
        result = self - ranges
        diff_generator = (
            IntRangeSet(tuple) - self for tuple in IntRangeSet._static_ranges(ranges)
        )
        result += diff_generator
        return result

    symmetric_difference = __xor__

    def _clone_state(self, result):
        self._start_items = result._start_items
        self._start_to_length = result._start_to_length
        return self

    def __iand__(*ranges_inputs):
        """
        Set the IntRangeSet to itself intersected with a input range

        These are the same:

        * ``a &= b``
        * ``a.intersection_update(b)``

        :Example:

        >>> a = IntRangeSet('0:5,6:11')
        >>> a &= '3:8'
        >>> print(a)
        IntRangeSet('3:5,6:8')
        """
        return ranges_inputs[0]._clone_state(IntRangeSet.intersection(*ranges_inputs))

    def intersection_update(*ranges_inputs):
        """See :meth:`IntRangeSet.__iand__`"""
        IntRangeSet.__iand__(*ranges_inputs)

    # same a-=b, a.difference_update(b,...), a.discard(b,...), a.remove(b,...). Note that "remove" is the only one that raises an error if the b,... aren't in a.
    # difference_update(other, ...)set -= other | ...
    # Update the set, removing elements found in others.
    # Changed in version 2.6: Accepts multiple input iterables.
    def __isub__(self, *ranges_inputs):
        """
        Remove the elements of the range inputs from the IntRangeSet

        These are the same:

        * ``a -= b``
        * ``a.difference_update(b)``
        * ``a.discard(b)``

        ``remove`` is almost the same except that it raises a KeyError if any element of b is not in a.

        * ``a.remove(b)``

        :Example:

        >>> a = IntRangeSet('0:5,6:11')
        >>> a -= '3:7'
        >>> print(a)
        IntRangeSet('0:3,7:11')


        The 'difference_update', 'discard' and 'remove' methods also support subtracting multiple ranges inputs.

        :Example:

        >>> a = IntRangeSet('0:5,6:11')
        >>> a.difference_update('3:7','8:100')
        >>> print(a)
        IntRangeSet('0:3,7')
        """
        for start, stop in IntRangeSet._static_ranges(*ranges_inputs):
            self._internal_isub(start, stop - start)
        return self

    def difference_update(self, *ranges_inputs):
        """See :meth:`IntRangeSet.__isub__`"""
        self.__isub__(*ranges_inputs)

    discard = difference_update

    def remove(self, *ranges_inputs):
        """See :meth:`IntRangeSet__isub__`"""
        for start_stop_tuple in IntRangeSet._static_ranges(*ranges_inputs):
            if start_stop_tuple not in self:
                raise KeyError()
            self -= start_stop_tuple

    def __ixor__(self, ranges):
        """
        Set the IntRangeSet to contains exactly those elements that appear in either itself or the input ranges but not both

        These are the same:

        * ``a ^= b``
        * ``a.symmetric_difference_update(b)``

        :Example:

        >>> a = IntRangeSet('0:5,6:11')
        >>> a ^= '3:8'
        >>> print(a)
        IntRangeSet('0:3,5,8:11')
        """
        return self._clone_state(self ^ ranges)

    def symmetric_difference_update(self, ranges):
        """See :meth:`IntRangeSet.__ixor__`"""
        self.__ixor__(ranges)

    # pop()
    def pop(self):
        """
        Remove and return the largest integer element from the IntRangeSet. Raises KeyError if the IntRangeSet is empty.

        :Example:

        >>> a = IntRangeSet('0:5,6:11')
        >>> print(a.pop())
        10
        >>> print(a)
        IntRangeSet('0:5,6:10')
        """
        if self.isempty:
            raise KeyError()
        # Get the last range
        start = self._start_items[-1]
        length = self._start_to_length[start]
        if length == 1:
            del self._start_to_length[start]
            self._start_items.pop()
        else:
            self._start_to_length[start] = length - 1
        return start + length - 1

    def __delitem__(self, key):
        """
        Remove elements from the IntRangeSet by position index. Position index can be specified by an integer with
        negative integers counting from the end. Position indexes can also be specified with slices and a ranges input.

        :Example:

        Removing with an integer position index:

        >>> a = IntRangeSet('100:200,1000')
        >>> del a[2]
        >>> print(a)
        IntRangeSet('100:102,103:200,1000')
        >>> del a[-1]
        >>> print(a)
        IntRangeSet('100:102,103:200')

        :Example:

        Removing with a slice:

        >>> a = IntRangeSet('100:200,1000')
        >>> del a[2:11]
        >>> print(a)
        IntRangeSet('100:102,111:200,1000')

        :Example:

        Removing with a ranges input:

        >>> a = IntRangeSet('100:200,1000')
        >>> del a['2:11']
        >>> print(a)
        IntRangeSet('100:102,111:200,1000')
        """
        if isinstance(key, numbers.Integral):
            if key >= 0:
                for start in self._start_items:
                    length = self._start_to_length[start]
                    if key < length:
                        self -= start + key
                        return
                    key -= length
                raise KeyError()
            else:
                assert key < 0
                key = -key - 1
                for start_index in range(len(self._start_items)):
                    start = self._start_items[-1 - start_index]
                    length = self._start_to_length[start]
                    if key < length:
                        self -= start + length - 1 - key
                        return
                    key -= length
                raise KeyError()
        elif isinstance(key, slice):
            lenx = len(self)
            start, stop, step = key.start, key.stop, key.step
            start = start or 0
            stop = stop or lenx
            step = step or 1

            if step == 1:
                self -= (self[start], self[stop - 1] + 1)
            else:
                self -= (self[index] for index in range(*key.indices(lenx)))
        else:
            start_and_stop_generator = (
                self._two_index(start_index, stop_index)
                for start_index, stop_index in IntRangeSet._static_ranges(key)
            )
            self -= start_and_stop_generator

    def _two_index(self, start_index, stop_index):
        start = self[start_index]
        if stop_index == start_index + 1:
            stop = start + 1
        else:
            stop = self[stop_index - 1] + 1
        return start, stop

    def __reversed__(self):
        """
        reversed(a) is a generator that produces the integer elements of a in order from largest to smallest.

        :Example:

        >>> for i in reversed(IntRangeSet('1:4,10')):
        ...     print(i)
        10
        3
        2
        1
        """
        for start in reversed(self._start_items):
            length = self._start_to_length[start]
            for item in range(start + length - 1, start - 1, -1):
                yield item

    @staticmethod
    # This will gather adjacent ranges together into one range, e.g.  1:4,4,5-6 -> 1-6
    def _static_ranges(*iterables):
        iter = IntRangeSet._inner_static_ranges(*iterables)
        try:
            start0, stop0 = next(iter)
            assert start0 < stop0, (
                "Invalid range. Start "
                + str(start0)
                + " must be less than stop "
                + str(stop0)
                + "."
            )
        except StopIteration:
            return
        while True:
            try:
                start1, stop1 = next(iter)
                assert start1 < stop1, (
                    "Invalid range. Start "
                    + str(start1)
                    + " must be less than stop "
                    + str(stop1)
                    + "."
                )
            except StopIteration:
                yield start0, stop0
                return
            if stop0 == start1:  # We don't try to merge all cases, just the most common
                stop0 = stop1
            elif stop1 == start0:
                start0 = start1
            else:
                yield start0, stop0
                start0, stop0 = start1, stop1

    @staticmethod
    def _inner_static_ranges(*iterables):
        for iterable in iterables:
            if isinstance(iterable, numbers.Integral):
                yield iterable, iterable + 1
            elif isinstance(iterable, tuple):
                assert (
                    len(iterable) == 2
                    and isinstance(iterable[0], numbers.Integral)
                    and isinstance(iterable[1], int)
                ), "Tuples must contain exactly two int elements that represent the start (inclusive) and stop (exclusive) elements of a range."
                yield iterable[0], iterable[1]
            elif isinstance(iterable, slice):
                start = iterable.start
                stop = iterable.stop
                step = iterable.step or 1
                assert (
                    start is not None
                    and start >= 0
                    and stop is not None
                    and start < stop
                    and step > 0
                ), "With slice, start and stop must be nonnegative numbers, stop must be more than start, and step, if given, must be at least 1"
                if step == 1:
                    yield start, stop
                else:
                    for start in range(start, stop, step):
                        yield start, start + 1
            elif iterable is None:
                pass
            elif isinstance(iterable, str):
                # Parses strings of the form -10:-4,-2-10,12-12. Spaces are allowed, no other characters are.
                #  will return an empty range
                if iterable == "":
                    pass
                else:
                    for range_string in iterable.split(","):
                        match = IntRangeSet._rangeExpression.match(
                            range_string
                        )  # LATER is there a good error message if it is not well formed?
                        if match is None:
                            raise Exception(
                                "The string is not well-formed. '{0}'".format(
                                    range_string
                                )
                            )
                        start = int(match.group("start"))
                        stop = int(match.group("stop") or start + 1)
                        yield start, stop
            elif hasattr(iterable, "ranges"):
                for start, stop in iterable.ranges():
                    yield start, stop
            elif hasattr(iterable, "__iter__"):
                for start, stop in IntRangeSet._static_ranges(*iterable):
                    yield start, stop
            else:
                raise Exception(
                    "Don't know how to construct a IntRangeSet from '{0}'".format(
                        iterable
                    )
                )

    def _internal_add(self, start, length=1):  # !! should it be "length" or "stop"
        assert length > 0, "Length must be greater than zero"
        assert len(self._start_items) == len(self._start_to_length)
        index = bisect_left(self._start_items, start)
        if index != len(self._start_items) and self._start_items[index] == start:
            if length <= self._start_to_length[start]:
                return
            else:
                self._start_to_length[start] = length
                index += 1  # index should point to the following range for the remainder of this method
                previous = start
                stop = start + length
        elif index == 0:
            self._start_items.insert(index, start)
            self._start_to_length[start] = length
            previous = start
            stop = start + length
            index += 1  # index_of_miss should point to the following range for the remainder of this method
        else:
            previous = self._start_items[index - 1]
            stop = previous + self._start_to_length[previous]

            if start <= stop:
                new_length = start - previous + length
                assert new_length > 0  # real assert
                if new_length < self._start_to_length[previous]:
                    return
                else:
                    self._start_to_length[previous] = new_length
                    stop = previous + new_length
            else:  # after previous range, not contiguous with previous range
                self._start_items.insert(index, start)
                self._start_to_length[start] = length
                previous = start
                stop = start + length
                index += 1

        if index == len(self._start_items):
            return

        # collapse next range into this one
        next = self._start_items[index]
        while stop >= next:
            new_stop = max(stop, next + self._start_to_length[next])
            self._start_to_length[previous] = (
                new_stop - previous
            )  # ItemToLength[previous] + ItemToLength[next]
            del self._start_to_length[next]
            del self._start_items[index]
            stop = new_stop
            if index >= len(self._start_items):
                break
            next = self._start_items[index]
        return

    # return the range that has the largest start and for which start<=element
    def _best_start_length_index_contains(self, element):
        index = bisect_left(self._start_items, element)
        if (
            index != len(self._start_items) and self._start_items[index] == element
        ):  # element is the start element of some range
            return element, self._start_to_length[element], index, True
        elif index == 0:  # element is before any of the ranges
            return element, 0, -1, False
        else:
            index -= 1
            start = self._start_items[index]
            length = self._start_to_length[start]
            return (
                start,
                length,
                index,
                element < start + length,
            )  # we already know element is greater than start

    def _delete_ranges(self, start_range_index, stop_range_index):
        for range_index in range(start_range_index, stop_range_index):
            del self._start_to_length[self._start_items[range_index]]
        del self._start_items[start_range_index:stop_range_index]

    def _shorten_last_range(self, stop_in, start1, length1, index1):
        delta_start = stop_in - start1
        self._start_items[index1] += delta_start
        del self._start_to_length[start1]
        self._start_to_length[start1 + delta_start] = length1 - delta_start
        assert len(self._start_items) == len(self._start_to_length)

    def _shorten_first_range(self, start_in, start0, length0):
        assert len(self._start_items) == len(self._start_to_length)
        self._start_to_length[start0] = start_in - start0
        assert len(self._start_items) == len(self._start_to_length)

    def _internal_isub(self, start_in, length_in=1):
        # !! should it be "length" or "stop"?
        assert length_in > 0, "Length must be greater than zero"
        assert len(self._start_items) == len(self._start_to_length)
        stop_in = start_in + length_in

        # return the range that has the largest start and for which start<=element
        start0, length0, index0, contains0 = self._best_start_length_index_contains(
            start_in
        )
        if length_in > 1:
            start1, length1, index1, contains1 = self._best_start_length_index_contains(
                stop_in - 1
            )
        else:
            start1, length1, index1, contains1 = start0, length0, index0, contains0

        # Is the front of first range unchanged, changed, or deleted?
        if not contains0:  # unchanged
            # Is the end of last range unchanged, changed, or deleted?
            if not contains1:  # unchanged
                self._delete_ranges(index0 + 1, index1 + 1)  # delete any middle range
            elif start1 + length1 == stop_in:  # deleted
                self._delete_ranges(
                    index0 + 1, index1 + 1
                )  # delete any middle and the last ranges
            else:  # changed
                assert index0 < index1, "real assert"
                self._shorten_last_range(
                    stop_in, start1, length1, index1
                )  # shorten last range
                self._delete_ranges(index0 + 1, index1)  # delete any middle ranges
        elif start0 == start_in:  # deleted
            # Is the end of last range unchanged, changed, or deleted?
            if not contains1:  # unchanged
                self._delete_ranges(
                    index0, index1 + 1
                )  # delete start range and any middle ranges
            elif start1 + length1 == stop_in:  # deleted
                self._delete_ranges(
                    index0, index1 + 1
                )  # delete start range and any middle ranges and last range
            else:  # changed
                assert index0 <= index1, "real assert"
                self._shorten_last_range(
                    stop_in, start1, length1, index1
                )  # shorten last range
                self._delete_ranges(
                    index0, index1
                )  # delete start range and any middle ranges
        else:  # changed
            # Is the end of last range unchanged, changed, or deleted?
            if not contains1:  # unchanged
                self._shorten_first_range(
                    start_in, start0, length0
                )  # shorten first range
                self._delete_ranges(index0 + 1, index1 + 1)  # delete any middle ranges
            elif start1 + length1 == stop_in:  # deleted
                self._shorten_first_range(
                    start_in, start0, length0
                )  # shorten first range
                self._delete_ranges(
                    index0 + 1, index1 + 1
                )  # delete any middle ranges and last range
            else:  # changed
                if index0 == index1:  # need to split into two ranges
                    self._start_items.insert(index0 + 1, stop_in)
                    self._start_to_length[stop_in] = start1 + length1 - stop_in
                    self._shorten_first_range(
                        start_in, start0, length0
                    )  # shorten first range
                else:
                    self._shorten_last_range(
                        stop_in, start1, length1, index1
                    )  # shorten last range
                    self._shorten_first_range(
                        start_in, start0, length0
                    )  # shorten first range
                    self._delete_ranges(index0 + 1, index1)  # delete any middle ranges
        assert len(self._start_items) == len(self._start_to_length)


class TestIntRangeSet(unittest.TestCase):
    def test_int_range_set(self):
        IntRangeSet._test()

    def test_doc(self):
        import pysnptools.util.intrangeset as set_mod

        doctest.testmod(set_mod)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    IntRangeSet._test()
    doctest.testmod(optionflags=doctest.ELLIPSIS)



from pysnptools.pstreader import PstNpz
from pysnptools.snpreader import SnpReader
import logging
import numpy as np
import warnings


class SnpNpz(PstNpz, SnpReader):
    """
    A :class:`.SnpReader` for reading \*.snp.npz files from disk.

    See :class:`.SnpReader` for general examples of using SnpReaders.

    The general NPZ format is described `here <http://docs.scipy.org/doc/numpy/reference/generated/numpy.savez.html>`__. The SnpNpz format stores
    val, iid, sid, and pos information in NPZ format.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The SnpNpz file to read.

        :Example:

        >>> from pysnptools.snpreader import SnpNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> npz_file = example_file('pysnptools/examples/toydata10.snp.npz')
        >>> data_on_disk = SnpNpz(npz_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 10)

    **Methods beyond** :class:`.SnpReader`

    """

    def __init__(self, *args, **kwargs):
        super(SnpNpz, self).__init__(*args, **kwargs)

    @property
    def row(self):
        self._run_once()
        if self._row.dtype.type is not np.str_:
            self._row = np.array(self._row, dtype="str")
        return self._row

    @property
    def col(self):
        self._run_once()
        if self._col.dtype.type is not np.str_:
            self._col = np.array(self._col, dtype="str")
        return self._col

    @staticmethod
    def write(filename, snpdata):
        """Writes a :class:`SnpData` to SnpNpz format and returns the :class:`.SnpNpz`

        :param filename: the name of the file to create
        :type filename: string
        :param snpdata: The in-memory data that should be written to disk.
        :type snpdata: :class:`SnpData`
        :rtype: :class:`.SnpNpz`

        >>> from pysnptools.snpreader import SnpNpz, Bed
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> snpdata = Bed(bed_file,count_A1=False)[:,:10].read()     # Read first 10 snps from Bed format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.snp.npz")
        >>> SnpNpz.write("tempdir/toydata10.snp.npz",snpdata)          # Write data in SnpNpz format
        SnpNpz('tempdir/toydata10.snp.npz')
        """
        row_ascii = np.array(
            snpdata.row, dtype="S"
        )  # !!! would be nice to avoid this copy when not needed.
        col_ascii = np.array(
            snpdata.col, dtype="S"
        )  # !!! would be nice to avoid this copy when not needed.
        np.savez(
            filename,
            row=row_ascii,
            col=col_ascii,
            row_property=snpdata.row_property,
            col_property=snpdata.col_property,
            val=snpdata.val,
        )
        logging.debug("Done writing " + filename)
        return SnpNpz(filename)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import logging
import os
import numpy as np
from pysnptools.pstreader import PstReader


# !!! would be better to make a Transpose class that could term _mergerows into mergecols? Be sure special Bed code is still there.
class _MergeCols(PstReader):
    def __init__(self, reader_list, cache_file=None, skip_check=False):
        super(_MergeCols, self).__init__()
        assert len(reader_list) > 0, "Expect at least one reader"
        self.skip_check = skip_check

        self.reader_list = reader_list
        self._repr_string = "_MergeCols({0})".format(
            ",".join([str(s) for s in reader_list])
        )

        if cache_file is not None:
            # !!!add warning if cache_file doesn't end with .npz
            if not os.path.exists(cache_file):
                self._run_once()
                self._savez(cache_file)
            else:
                self._load(cache_file)

        def _savez(self, cache_file):
            np.savez(
                cache_file,
                _row=self._row,
                _row_property=self._row_property,
                _col=self._col,
                _col_property=self._col_property,
                col_count_list=self.col_count_list,
            )

        def _load(self, cache_file):
            with np.load(cache_file, allow_pickle=True) as data:
                self._col = data["_col"]
                self._col_property = data["_col_property"]
                self.col_count_list = data["col_count_list"]
                assert ("_row" in data) == ("_row_property" in data)
                self._row = data["_row"]
                self._row_property = data["_row_property"]

    def _run_once(self):
        if hasattr(self, "_has_run_once"):
            return
        from pysnptools.snpreader import Bed

        self._has_run_once = True
        # Check that all rows are distinct and that all cols and pos are the same and in the same order

        if self.skip_check and all(
            isinstance(reader, Bed) for reader in self.reader_list
        ):  # Special code if all Bed readers
            l = [
                SnpReader._read_map_or_bim(
                    reader.filename, remove_suffix="bed", add_suffix="bim"
                )
                for reader in self.reader_list
            ]
            self.col_count_list = np.array([len(ll[0]) for ll in l])
            self._row = self.reader_list[0].row
            self._row_property = self.reader_list[0].row_property
            self._col = np.concatenate([ll[0] for ll in l])
            self._col_property = np.concatenate([ll[1] for ll in l])
        else:
            col_list = []
            col_property_list = []
            col_set = set()
            self.col_count_list = []
            self._row = self.reader_list[0].row
            self._row_property = self.reader_list[0].row_property
            for reader_index, reader in enumerate(self.reader_list):
                if reader_index % 10 == 0:
                    logging.info(
                        "_MergeCols looking at reader #{0}: {1}".format(
                            reader_index, reader
                        )
                    )
                if not self.skip_check:
                    assert np.array_equal(
                        self._row, reader.row
                    ), "Expect rows to be the same across all files"
                    np.testing.assert_equal(
                        self._row_property, reader.row_property
                    )  # "Expect column_property to be the same across all files"
                    size_before = len(col_set)
                    col_set.update((tuple(item) for item in reader.col))
                    assert (
                        len(col_set) == size_before + reader.col_count
                    ), "Expect cols to be distinct in all files"
                col_list.append(reader.col)
                col_property_list.append(reader.col_property)
                self.col_count_list.append(reader.col_count)
            self._col = np.concatenate(col_list)
            self._col_property = np.concatenate(col_property_list)
            self.col_count_list = np.array(self.col_count_list)

    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    @property
    def row_property(self):
        self._run_once()
        return self._row_property

    def __repr__(self):
        # Don't need _run_once because based only on initial info
        return self._repr_string

    def copyinputs(self, copier):
        self._run_once()
        for reader in self.reader_list:
            copier.input(reader)

    def _create_reader_and_col_index_list(self, col_index):
        result = []
        start = 0
        for reader_index in range(len(self.reader_list)):
            stop = (
                start + self.col_count_list[reader_index]
            )  # !!! shouldn't this be col_count (and check _mergerows, too)
            is_here = (col_index >= start) * (col_index < stop)
            if any(is_here):
                col_index_rel = col_index[is_here] - start
                result.append((reader_index, is_here, col_index_rel))
            start = stop
        return result

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        # !!!tests to do: no row's
        # !!!tests to do: no col's
        # !!!test to do: from file 1, file2, and then file1 again
        dtype = np.dtype(dtype)
        col_index = (
            col_index_or_none
            if col_index_or_none is not None
            else np.arange(self.col_count)
        )  # !!!might want to special case reading all
        row_index_or_none_count = (
            self.row_count if row_index_or_none is None else len(row_index_or_none)
        )
        col_index_or_none_count = (
            self.col_count if col_index_or_none is None else len(col_index_or_none)
        )

        # Create a list of (reader,col_index)
        reader_and_col_index_list = self._create_reader_and_col_index_list(col_index)
        if len(reader_and_col_index_list) == 0:
            return self.reader_list[0]._read(
                row_index_or_none,
                col_index,
                order,
                dtype,
                force_python_only,
                view_ok,
                num_threads,
            )
        elif len(reader_and_col_index_list) == 1:
            reader_index, col_index_in, col_index_rel = reader_and_col_index_list[0]
            reader = self.reader_list[reader_index]
            return reader._read(
                row_index_or_none,
                col_index_rel,
                order,
                dtype,
                force_python_only,
                view_ok,
                num_threads,
            )
        else:
            logging.info(
                "Starting read from {0} subreaders".format(
                    len(reader_and_col_index_list)
                )
            )
            if (
                order == "A" or order is None
            ):  # LATER does every _read( need code like this?
                order = "F"
            val = None

            for reader_index, is_here, col_index_rel in reader_and_col_index_list:
                reader = self.reader_list[reader_index]
                if reader_index % 1 == 0:
                    logging.info("Reading from #{0}: {1}".format(reader_index, reader))
                piece = reader._read(
                    row_index_or_none,
                    col_index_rel,
                    order,
                    dtype,
                    force_python_only,
                    view_ok=True,
                    num_threads=num_threads,
                )
                if val is None:
                    if len(piece.shape) == 2:
                        val = np.empty(
                            [row_index_or_none_count, col_index_or_none_count],
                            dtype=dtype,
                            order=order,
                        )
                    else:
                        val = np.empty(
                            [
                                row_index_or_none_count,
                                col_index_or_none_count,
                                self.val_shape,
                            ],
                            dtype=dtype,
                            order=order,
                        )
                    val.fill(np.nan)  # LATER Keep this?

                if len(val.shape) == 2:
                    val[:, is_here] = piece
                else:
                    val[:, is_here, :] = piece

            logging.info(
                "Ended read from {0} subreaders".format(len(reader_and_col_index_list))
            )
            assert val is not None
            return val



import numpy as np
import scipy as sp
import logging
import doctest
import shutil
import unittest
import os.path
import time
import sys

from pysnptools.distreader.distmemmap import TestDistMemMap

try:
    from bgen_reader.test.test_bgen_reader import nowrite_permission, noread_permission

    BGEN_READER_AVAILABLE = True
except ImportError:
    BGEN_READER_AVAILABLE = False
from pysnptools.distreader.distgen import TestDistGen
from pysnptools.distreader import (
    DistNpz,
    DistHdf5,
    DistMemMap,
    DistData,
    _DistMergeSIDs,
)
from pysnptools.util import create_directory_if_necessary
from pysnptools.snpreader import SnpNpz, Bed
from pysnptools.kernelreader.test import _fortesting_JustCheckExists
from pysnptools.pstreader import PstData
from pysnptools.snpreader import SnpData


class TestDistReaders(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.currentFolder = os.path.dirname(os.path.realpath(__file__))
        # TODO: get data set with NANs!
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        self.distdata = distreader.read(order="F", force_python_only=True)
        self.dist_values = self.distdata.val

    def test_3d(self):
        from pysnptools.distreader import DistData

        np.random.seed(0)
        row_count = 4
        col_count = 5
        val_shape = 3
        val = np.random.random((row_count, col_count, val_shape))
        val /= val.sum(axis=2, keepdims=True)  # make probabilities sum to 1
        distdata = DistData(
            val=val,
            iid=[["iid{0}".format(i)] * 2 for i in range(row_count)],
            sid=["sid{0}".format(s) for s in range(col_count)],
        )

    def test_scalar_index(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        arr = np.int64(1)
        distreader[arr, arr]

    def test_c_reader_hdf5(self):
        distreader = DistHdf5(
            self.currentFolder + "/../examples/toydata.snpmajor.dist.hdf5"
        )
        self.c_reader(distreader)

    def test_hdf5_case3(self):
        distreader1 = DistHdf5(
            self.currentFolder + "/../examples/toydata.snpmajor.dist.hdf5"
        )[::2, :]
        distreader2 = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")[
            ::2, :
        ]
        self.assertTrue(
            np.allclose(
                distreader1.read().val, distreader2.read().val, rtol=1e-05, atol=1e-05
            )
        )

    def test_c_reader_npz(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata10.dist.npz")
        distdata = distreader.read(order="F", force_python_only=False)
        snp_c = distdata.val

        self.assertEqual(np.float64, snp_c.dtype)
        self.assertTrue(
            np.allclose(self.dist_values[:, :10], snp_c, rtol=1e-05, atol=1e-05)
        )

        distreader1 = DistNpz(self.currentFolder + "/../examples/toydata10.dist.npz")
        distreader2 = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")[
            :, :10
        ]
        self.assertTrue(
            np.allclose(
                distreader1.read().val, distreader2.read().val, rtol=1e-05, atol=1e-05
            )
        )

        distdata.val[1, 2] = (
            np.NaN
        )  # Inject a missing value to test writing and reading missing values
        output = "tempdir/distreader/toydata10.dist.npz"
        create_directory_if_necessary(output)
        DistNpz.write(output, distdata)
        snpdata2 = DistNpz(output).read()
        np.testing.assert_array_almost_equal(distdata.val, snpdata2.val, decimal=10)

        snpdata3 = distdata[:, 0:0].read()  # create distdata with no sids
        output = "tempdir/distreader/toydata0.dist.npz"
        DistNpz.write(output, snpdata3)
        snpdata4 = DistNpz(output).read()
        assert snpdata3 == snpdata4

    def c_reader(self, distreader):
        """
        make sure c-reader yields same result
        """
        distdata = distreader.read(order="F", force_python_only=False)
        snp_c = distdata.val

        self.assertEqual(np.float64, snp_c.dtype)
        self.assertTrue(np.allclose(self.dist_values, snp_c, rtol=1e-05, atol=1e-05))
        return distdata

    def test_write_distnpz_f64cpp_0(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        iid_index = 0
        logging.info("iid={0}".format(iid_index))
        # if distreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    distreader = distreader[0:-1,:]
        # assert distreader.iid_count % 4 != 0
        distdata = distreader[0:iid_index, :].read(order="F", dtype=np.float64)
        if distdata.iid_count > 0:
            distdata.val[-1, 0] = float("NAN")
        output = "tempdir/toydata.F64cpp.{0}.dist.npz".format(iid_index)
        create_directory_if_necessary(output)
        DistNpz.write(output, distdata)
        snpdata2 = DistNpz(output).read()
        np.testing.assert_array_almost_equal(distdata.val, snpdata2.val, decimal=10)

    def test_write_distnpz_f64cpp_1(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        iid_index = 1
        logging.info("iid={0}".format(iid_index))
        # if distreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    distreader = distreader[0:-1,:]
        # assert distreader.iid_count % 4 != 0
        distdata = distreader[0:iid_index, :].read(order="F", dtype=np.float64)
        if distdata.iid_count > 0:
            distdata.val[-1, 0] = float("NAN")
        output = "tempdir/toydata.F64cpp.{0}.dist.npz".format(iid_index)
        create_directory_if_necessary(output)
        DistNpz.write(output, distdata)
        snpdata2 = DistNpz(output).read()
        np.testing.assert_array_almost_equal(distdata.val, snpdata2.val, decimal=10)

    def test_write_distnpz_f64cpp_5(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")

        _fortesting_JustCheckExists().input(distreader)

        iid_index = 5
        logging.info("iid={0}".format(iid_index))
        # if distreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    distreader = distreader[0:-1,:]
        # assert distreader.iid_count % 4 != 0
        distdata = distreader[0:iid_index, :].read(order="F", dtype=np.float64)
        if distdata.iid_count > 0:
            distdata.val[-1, 0] = float("NAN")
        output = "tempdir/toydata.F64cpp.{0}.dist.npz".format(iid_index)
        create_directory_if_necessary(output)
        DistNpz.write(output, distdata)  # ,force_python_only=True)
        snpdata2 = DistNpz(output).read()
        np.testing.assert_array_almost_equal(distdata.val, snpdata2.val, decimal=10)

    def test_write_x_x_cpp(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        for order in ["C", "F"]:
            for dtype in [np.float32, np.float64]:
                distdata = distreader.read(order=order, dtype=dtype)
                distdata.val[-1, 0] = float("NAN")
                output = "tempdir/toydata.{0}{1}.cpp.dist.npz".format(
                    order, "32" if dtype == np.float32 else "64"
                )
                create_directory_if_necessary(output)
                DistNpz.write(output, distdata)
                snpdata2 = DistNpz(output).read()
                np.testing.assert_array_almost_equal(
                    distdata.val, snpdata2.val, decimal=10
                )

    def test_subset_view(self):
        distreader2 = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")[
            :, :
        ]
        result = distreader2.read(view_ok=True)
        self.assertFalse(distreader2 is result)
        result2 = result[:, :].read()
        self.assertFalse(np.may_share_memory(result2.val, result.val))
        result3 = result[:, :].read(view_ok=True)
        self.assertTrue(np.may_share_memory(result3.val, result.val))
        result4 = result3.read()
        self.assertFalse(np.may_share_memory(result4.val, result3.val))
        result5 = result4.read(view_ok=True)
        self.assertTrue(np.may_share_memory(result4.val, result5.val))

    def test_writes(self):
        from pysnptools.distreader import DistData, DistHdf5, DistNpz, DistMemMap
        from pysnptools.kernelreader.test import _fortesting_JustCheckExists

        the_class_and_suffix_list = [
            (DistNpz, "npz", None, None),
            # Bgen used to be here
            (DistHdf5, "hdf5", None, None),
            (DistMemMap, "memmap", None, None),
        ]
        # Skip because write requires qctool, which is not installed by default
        #     from pysnptools.distreader import Bgen
        #     the_class_and_suffix_list.append((Bgen,"bgen",None,lambda filename,distdata: Bgen.write(filename,distdata,bits=32)))

        cant_do_col_prop_none_set = {"bgen"}
        cant_do_col_len_0_set = {"bgen"}
        cant_do_row_count_zero_set = {"bgen"}
        can_swap_0_2_set = {}
        can_change_col_names_set = {}
        ignore_fam_id_set = {}
        ignore_pos1_set = {"bgen"}
        ignore_pos_set = {}
        erase_any_write_dir = {}

        # ===================================
        #    Starting main function
        # ===================================
        logging.info("starting 'test_writes'")
        np.random.seed(0)
        output_template = "tempdir/distreader/writes.{0}.{1}"
        create_directory_if_necessary(output_template.format(0, "npz"))
        i = 0
        for row_count in [0, 5, 2, 1]:
            for col_count in [4, 2, 1, 0]:
                val = np.random.random(size=[row_count, col_count, 3])
                val /= val.sum(axis=2, keepdims=True)  # make probabilities sum to 1

                val[val == 3] = np.NaN
                row = [("0", "0"), ("1", "1"), ("2", "2"), ("3", "3"), ("4", "4")][
                    :row_count
                ]
                col = ["s0", "s1", "s2", "s3", "s4"][:col_count]
                for is_none in [True, False]:
                    row_prop = None
                    col_prop = (
                        None if is_none else [(x, x, x) for x in range(5)][:col_count]
                    )
                    distdata = DistData(
                        iid=row, sid=col, val=val, pos=col_prop, name=str(i)
                    )
                    for (
                        the_class,
                        suffix,
                        constructor,
                        writer,
                    ) in the_class_and_suffix_list:
                        constructor = constructor or (
                            lambda filename: the_class(filename)
                        )
                        writer = writer or (
                            lambda filename, distdata: the_class.write(
                                filename, distdata
                            )
                        )
                        if col_count == 0 and suffix in cant_do_col_len_0_set:
                            continue
                        if col_prop is None and suffix in cant_do_col_prop_none_set:
                            continue
                        if row_count == 0 and suffix in cant_do_row_count_zero_set:
                            continue
                        filename = output_template.format(i, suffix)
                        logging.info(filename)
                        i += 1
                        if suffix in erase_any_write_dir and os.path.exists(filename):
                            shutil.rmtree(filename)
                        ret = writer(filename, distdata)
                        assert ret is not None
                        for subsetter in [None, np.s_[::2, ::3]]:
                            reader = constructor(filename)
                            _fortesting_JustCheckExists().input(reader)
                            subreader = (
                                reader
                                if subsetter is None
                                else reader[subsetter[0], subsetter[1]]
                            )
                            readdata = subreader.read(order="C")
                            expected = (
                                distdata
                                if subsetter is None
                                else distdata[subsetter[0], subsetter[1]].read()
                            )
                            if not suffix in can_swap_0_2_set:
                                assert np.allclose(
                                    readdata.val, expected.val, equal_nan=True
                                )
                            else:
                                for col_index in range(readdata.col_count):
                                    assert np.allclose(
                                        readdata.val[:, col_index],
                                        expected.val[:, col_index],
                                        equal_nan=True,
                                    ) or np.allclose(
                                        readdata.val[:, col_index] * -1 + 2,
                                        expected.val[:, col_index],
                                        equal_nan=True,
                                    )
                            if not suffix in ignore_fam_id_set:
                                assert np.array_equal(readdata.row, expected.row)
                            else:
                                assert np.array_equal(
                                    readdata.row[:, 1], expected.row[:, 1]
                                )
                            if not suffix in can_change_col_names_set:
                                assert np.array_equal(readdata.col, expected.col)
                            else:
                                assert readdata.col_count == expected.col_count
                            assert np.array_equal(
                                readdata.row_property, expected.row_property
                            ) or (
                                readdata.row_property.shape[1] == 0
                                and expected.row_property.shape[1] == 0
                            )

                            if suffix in ignore_pos1_set:
                                assert np.allclose(
                                    readdata.col_property[:, [0, 2]],
                                    expected.col_property[:, [0, 2]],
                                    equal_nan=True,
                                ) or (
                                    readdata.col_property.shape[1] == 0
                                    and expected.col_property.shape[1] == 0
                                )
                            elif not suffix in ignore_pos_set:
                                assert np.allclose(
                                    readdata.col_property,
                                    expected.col_property,
                                    equal_nan=True,
                                ) or (
                                    readdata.col_property.shape[1] == 0
                                    and expected.col_property.shape[1] == 0
                                )
                            else:
                                assert len(readdata.col_property) == len(
                                    expected.col_property
                                )
                        try:
                            os.remove(filename)
                        except Exception:
                            pass
        logging.info("done with 'test_writes'")

    def test_block_size_DistData(self):
        np.random.seed(0)
        sid_count = 20
        val = np.array(
            np.random.random(size=[3, sid_count, 3]), dtype=np.float64, order="F"
        )
        val /= val.sum(axis=2, keepdims=True)  # make probabilities sum to 1
        distreader = DistData(
            iid=[["0", "0"], ["1", "1"], ["2", "2"]],
            sid=[str(i) for i in range(sid_count)],
            val=val,
        )
        snpdata0 = distreader.as_snp(max_weight=100, block_size=1).read()
        snpdata1 = distreader.as_snp(max_weight=100, block_size=None).read()
        np.testing.assert_array_almost_equal(snpdata0.val, snpdata1.val, decimal=10)

    def test_block_size_Snp2Dist(self):
        from pysnptools.snpreader import SnpData
        from pysnptools.distreader._snp2dist import _Snp2Dist

        np.random.seed(0)
        sid_count = 20
        val = np.array(
            np.random.randint(0, 3, size=[3, sid_count]), dtype=np.float64, order="F"
        )
        snpreader = SnpData(
            iid=[["0", "0"], ["1", "1"], ["2", "2"]],
            sid=[str(i) for i in range(sid_count)],
            val=val,
        )
        distdata0 = snpreader.as_dist(max_weight=2, block_size=1).read()
        distdata1 = snpreader.as_dist(max_weight=2, block_size=None).read()
        np.testing.assert_array_almost_equal(distdata0.val, distdata1.val, decimal=10)

    def test_intersection_Dist2Snp(self):
        from pysnptools.snpreader._dist2snp import _Dist2Snp
        from pysnptools.snpreader import Pheno
        from pysnptools.distreader._subset import _DistSubset
        from pysnptools.snpreader._subset import _SnpSubset
        from pysnptools.util import intersect_apply

        dist_all = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        k = dist_all.as_snp(max_weight=25)

        pheno = Pheno(self.currentFolder + "/../examples/toydata.phe")
        pheno = pheno[1:, :]  # To test intersection we remove a iid from pheno

        k1, pheno = intersect_apply([k, pheno])
        assert isinstance(k1.distreader, _DistSubset) and not isinstance(k1, _SnpSubset)

        # What happens with fancy selection?
        k2 = k[::2, :]
        assert isinstance(k2, _Dist2Snp)

        logging.info("Done with test_intersection")

    def test_intersection_Snp2Dist(self):
        from pysnptools.distreader._snp2dist import _Snp2Dist
        from pysnptools.snpreader import Pheno, Bed
        from pysnptools.distreader._subset import _DistSubset
        from pysnptools.snpreader._subset import _SnpSubset
        from pysnptools.util import intersect_apply

        snp_all = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=True
        )
        k = snp_all.as_dist(max_weight=2)

        pheno = Pheno(self.currentFolder + "/../examples/toydata.phe")
        pheno = pheno[1:, :]  # To test intersection we remove a iid from pheno

        k1, pheno = intersect_apply([k, pheno])
        assert isinstance(k1.snpreader, _SnpSubset) and not isinstance(k1, _DistSubset)

        # What happens with fancy selection?
        k2 = k[::2, :]
        assert isinstance(k2, _Snp2Dist)

        logging.info("Done with test_intersection")

    def test_roundtrip(self):
        max_weight = 2
        snpreader1 = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=True
        )
        snpdata1 = snpreader1.read()
        distreader1 = snpreader1.as_dist(max_weight)
        snpreader2 = distreader1.as_snp(max_weight)
        assert snpdata1.allclose(snpreader2.read(), equal_nan=True)
        snpdata1.val[0, 0] = np.nan
        assert snpdata1.allclose(
            snpdata1.as_dist(max_weight).as_snp(max_weight).read(), equal_nan=True
        )

    def test_dist_snp2(self):
        logging.info("in test_dist_snp2")
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        dist2snp = distreader.as_snp(max_weight=33)
        s = str(dist2snp)
        _fortesting_JustCheckExists().input(dist2snp)

    def test_snp_dist2(self):
        logging.info("in test_snp_dist2")
        snpreader = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        snp2dist = snpreader.as_dist(max_weight=2)
        s = str(snp2dist)
        _fortesting_JustCheckExists().input(snp2dist)

    def test_subset_Dist2Snp(self):
        logging.info("in test_subset")
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        dist2snp = distreader.as_snp(max_weight=10)
        dssub = dist2snp[::2, ::2]
        snpdata1 = dssub.read()
        expected = distreader.as_snp(max_weight=10)[::2, ::2].read()
        np.testing.assert_array_almost_equal(snpdata1.val, expected.val, decimal=10)

        logging.info("done with test")

    def test_subset_Snp2Dist(self):  # !!!move these to another test class
        logging.info("in test_subset")
        snpreader = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        snp2dist = snpreader.as_dist(max_weight=2)
        sub = snp2dist[::2, ::2]
        distdata1 = sub.read()
        expected = snpreader.as_dist(max_weight=2)[::2, ::2].read()
        np.testing.assert_array_almost_equal(distdata1.val, expected.val, decimal=10)

        logging.info("done with test")

    def test_respect_inputs_DistData(self):
        np.random.seed(0)
        for dtype_start, decimal_start in [(np.float32, 5), (np.float64, 10)]:
            for order_start in ["F", "C", "A"]:
                for sid_count in [20, 2]:
                    val = np.array(
                        np.random.random(size=[3, sid_count, 3]),
                        dtype=dtype_start,
                        order=order_start,
                    )
                    val /= val.sum(axis=2, keepdims=True)  # make probabilities sum to 1
                    distdataX = DistData(
                        iid=[["0", "0"], ["1", "1"], ["2", "2"]],
                        sid=[str(i) for i in range(sid_count)],
                        val=val,
                    )
                    for max_weight in [1.0, 2.0]:
                        weights = np.array([0, 0.5, 1]) * max_weight
                        for distreader0 in [distdataX, distdataX[:, 1:]]:
                            distreader1 = distreader0[1:, :]
                            refdata0 = distreader0.read()
                            refval0 = (refdata0.val * weights).sum(axis=-1)
                            for dtype_goal, decimal_goal in [
                                (np.float32, 5),
                                (np.float64, 10),
                            ]:
                                for order_goal in ["F", "C", "A"]:
                                    k = distreader0.as_snp(
                                        max_weight=max_weight, block_size=1
                                    ).read(order=order_goal, dtype=dtype_goal)
                                    DistData._array_properties_are_ok(
                                        k.val, order_goal, dtype_goal
                                    )
                                    np.testing.assert_array_almost_equal(
                                        refval0,
                                        k.val,
                                        decimal=min(decimal_start, decimal_goal),
                                    )

    def test_npz(self):
        logging.info("in test_npz")
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        snpdata1 = distreader.as_snp(max_weight=1.0).read()
        s = str(snpdata1)
        output = "tempdir/distreader/toydata.snp.npz"
        create_directory_if_necessary(output)
        SnpNpz.write(output, snpdata1)
        snpreader2 = SnpNpz(output)
        snpdata2 = snpreader2.read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata2.val, decimal=10)
        logging.info("done with test")


class TestDistNaNCNC(unittest.TestCase):
    def __init__(
        self,
        iid_index_list,
        snp_index_list,
        distreader,
        dtype,
        order,
        force_python_only,
        reference_snps,
        reference_dtype,
    ):
        self.iid_index_list = iid_index_list
        self.snp_index_list = snp_index_list
        self.distreader = distreader
        self.dtype = np.dtype(dtype)
        self.order = order
        self.force_python_only = force_python_only
        self.reference_snps = reference_snps
        self.reference_dtype = reference_dtype

    _testMethodName = "runTest"
    _testMethodDoc = None

    @staticmethod
    def factory_iterator():
        snp_reader_factory_distnpz = lambda: DistNpz("../examples/toydata.dist.npz")
        snp_reader_factory_snpmajor_hdf5 = lambda: DistHdf5(
            "../examples/toydata.snpmajor.dist.hdf5"
        )
        snp_reader_factory_iidmajor_hdf5 = lambda: DistHdf5(
            "../examples/toydata.iidmajor.dist.hdf5"
        )

        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        distreader0 = snp_reader_factory_distnpz()
        S_original = distreader0.sid_count
        N_original = distreader0.iid_count

        snps_to_read_count = min(S_original, 100)

        for iid_index_list in [
            range(N_original),
            range(N_original // 2),
            range(N_original - 1, 0, -2),
        ]:
            for snp_index_list in [
                range(snps_to_read_count),
                range(snps_to_read_count // 2),
                range(snps_to_read_count - 1, 0, -2),
            ]:
                reference_snps, reference_dtype = TestDistNaNCNC(
                    iid_index_list,
                    snp_index_list,
                    snp_reader_factory_distnpz(),
                    np.float64,
                    "C",
                    "False",
                    None,
                    None,
                ).read_and_standardize()
                for distreader_factory in [
                    snp_reader_factory_distnpz,
                    snp_reader_factory_snpmajor_hdf5,
                    snp_reader_factory_iidmajor_hdf5,
                ]:
                    for dtype in [np.float64, np.float32]:
                        for order in ["C", "F"]:
                            for force_python_only in [False, True]:
                                distreader = distreader_factory()
                                test_case = TestDistNaNCNC(
                                    iid_index_list,
                                    snp_index_list,
                                    distreader,
                                    dtype,
                                    order,
                                    force_python_only,
                                    reference_snps,
                                    reference_dtype,
                                )
                                yield test_case
        os.chdir(previous_wd)

    def __str__(self):
        iid_index_list = self.iid_index_list
        snp_index_list = self.snp_index_list
        distreader = self.distreader
        dtype = self.dtype
        order = self.order
        force_python_only = self.force_python_only
        return "{0}(iid_index_list=[{1}], snp_index_list=[{2}], distreader={3}, dtype={4}, order='{5}', force_python_only=={6})".format(
            self.__class__.__name__,
            (
                ",".join([str(i) for i in iid_index_list])
                if len(iid_index_list) < 10
                else ",".join([str(i) for i in iid_index_list[0:10]]) + ",..."
            ),
            (
                ",".join([str(i) for i in snp_index_list])
                if len(snp_index_list) < 10
                else ",".join([str(i) for i in snp_index_list[0:10]]) + ",..."
            ),
            distreader,
            dtype,
            order,
            force_python_only,
        )

    def read_and_standardize(self):
        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        iid_index_list = self.iid_index_list
        snp_index_list = self.snp_index_list
        distreader = self.distreader
        dtype = self.dtype
        order = self.order
        force_python_only = self.force_python_only

        snps = (
            distreader[iid_index_list, snp_index_list]
            .read(order=order, dtype=dtype, force_python_only=force_python_only)
            .val
        )
        snps[0, 0] = [np.nan] * 3  # add a NaN
        snps[:, 1] = [0.1, 0.2, 0.7]  # make a SNC
        os.chdir(previous_wd)
        return snps, dtype

    def doCleanups(self):
        pass
        # return super(NaNCNCTestCases, self).doCleanups()

    def runTest(self, result=None):
        snps, dtype = self.read_and_standardize()
        assert not np.array_equal(
            snps[0, 0], snps[0, 0]
        )  # without SnpReader's standardization NaN's stay NaN's
        assert np.allclose(snps[0, 1], [0.1, 0.2, 0.7])
        if self.reference_snps is not None:
            self.assertTrue(
                np.allclose(
                    self.reference_snps,
                    snps,
                    rtol=(
                        1e-04
                        if dtype == np.float32 or self.reference_dtype == np.float32
                        else 1e-12
                    ),
                    equal_nan=True,
                )
            )


# We do it this way instead of using doctest.DocTestSuite because doctest.DocTestSuite requires modules to be pickled, which python doesn't allow.
# We need tests to be pickleable so that they can be run on a cluster.
class TestDistReaderDocStrings(unittest.TestCase):
    def test_distreader(self):
        import pysnptools.distreader.distreader

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.distreader.distreader, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_distdata(self):
        import pysnptools.distreader.distdata

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.distreader.distdata, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_disthdf5(self):
        import pysnptools.distreader.disthdf5

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.distreader.disthdf5, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_distnpz(self):
        import pysnptools.distreader.distnpz

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.distreader.distnpz, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])

    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestBgen))
    test_suite.addTests(
        unittest.TestLoader().loadTestsFromTestCase(TestDistReaderDocStrings)
    )
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistGen))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistMemMap))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistReaders))
    test_suite.addTests(TestDistNaNCNC.factory_iterator())

    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.WARN)

    if False:
        from pysnptools.snpreader import Bed
        from pysnptools.distreader import DistData, DistNpz

        # Create toydata.dist.npz
        currentFolder = os.path.dirname(os.path.realpath(__file__))
        if True:
            snpreader = Bed(
                currentFolder + "/../examples/toydata.5chrom.bed", count_A1=True
            )[:25, :]
            np.random.seed(392)
            val = np.random.random((snpreader.iid_count, snpreader.sid_count, 3))
            val /= val.sum(axis=2, keepdims=True)  # make probabilities sum to 1
            distdata = DistData(
                iid=snpreader.iid, sid=snpreader.sid, pos=snpreader.pos, val=val
            )
            DistNpz.write(currentFolder + "/../examples/toydata.dist.npz", distdata)
        if True:
            distdata = DistNpz(currentFolder + "/../examples/toydata.dist.npz").read()
            for sid_major, name_bit in [(False, "iidmajor"), (True, "snpmajor")]:
                DistHdf5.write(
                    currentFolder
                    + "/../examples/toydata.{0}.dist.hdf5".format(name_bit),
                    distdata,
                    sid_major=sid_major,
                )
        if True:
            distdata = DistNpz(currentFolder + "/../examples/toydata.dist.npz")[
                :, :10
            ].read()
            DistNpz.write(currentFolder + "/../examples/toydata10.dist.npz", distdata)
        if True:
            distdata = DistNpz(currentFolder + "/../examples/toydata.dist.npz")[
                :, :10
            ].read()
            DistMemMap.write(currentFolder + "/../examples/tiny.dist.memmap", distdata)
        print("done")

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()



import numpy as np
from pysnptools.snpreader import SnpReader
from pysnptools.pstreader import _MergeCols

class _MergeSIDs(_MergeCols,SnpReader):
    def __init__(self,  *args, **kwargs):
        super(_MergeSIDs, self).__init__(*args, **kwargs)

    def _savez(self, cache_file):
        np.savez(cache_file,
                    _row=np.array(self._row,dtype='S'), _row_property=self._row_property,
                    _col=np.array(self._col,dtype='S'), _col_property=self._col_property,
                    sid_count_list=self.col_count_list)

    def _load(self,cache_file):
            with np.load(cache_file,allow_pickle=True) as data:
                self._col = np.array(data['_col'],dtype='str')
                self._col_property = data['_col_property']
                self.col_count_list = data['sid_count_list']
                assert ('_row' in data) == ('_row_property' in data)
                self._row = np.array(data['_row'],dtype='str')
                self._row_property = data['_row_property']



import logging
import pysnptools.test
import unittest


if __name__ == "__main__":
    logging.basicConfig(level=logging.WARN)

    suites = unittest.TestSuite([pysnptools.test.getTestSuite()])
    suites.debug

    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    logging.info("done with testing")



import numpy as np
import logging

from pysnptools.standardizer.standardizer import Standardizer
from pysnptools.standardizer.beta import Beta
from pysnptools.standardizer.unit import Unit
from pysnptools.standardizer.identity import Identity
from pysnptools.standardizer.diag_K_to_N import DiagKtoN
from pysnptools.standardizer.betatrained import BetaTrained
from pysnptools.standardizer.unittrained import UnitTrained
from pysnptools.standardizer.diag_K_to_N import DiagKtoNTrained

#Deprecated
from pysnptools.standardizer.bysqrtsidcount import BySqrtSidCount
from pysnptools.standardizer.bysidcount import BySidCount


import logging
import numpy as np
import unittest
import os
import shutil
import doctest
import pysnptools.util as pstutil
from pysnptools.pstreader import PstReader, PstData

_magic_number = 22891


class PstMemMap(PstData):
    """
    A :class:`.PstData` that keeps its data in a memory-mapped file. This allows data large than fits in main memory.

    See :class:`.PstData` for general examples of using PstData.

    **Constructor:**
        :Parameters: **filename** (*string*) -- The *\*.pst.memmap* file to read.

        Also see :meth:`.PstMemMap.empty` and :meth:`.PstMemMap.write`.

        :Example:

        >>> from pysnptools.pstreader import PstMemMap
        >>> from pysnptools.util import example_file # Download and return local file name
        >>>
        >>> pst_mem_map_file = example_file('pysnptools/examples/tiny.pst.memmap')
        >>> pst_mem_map = PstMemMap(pst_mem_map_file)
        >>> print(pst_mem_map.val[0,1], pst_mem_map.row_count, pst_mem_map.col_count)
        2.0 3 2

    **Methods beyond** :class:`PstReader`

    """

    def __init__(self, filename):
        PstReader.__init__(self)
        self._ran_once = False
        self._filename = filename

    def __repr__(self):
        return "{0}('{1}')".format(self.__class__.__name__, self._filename)

    def __getstate__(self):
        return self.filename

    def __setstate__(self, state):
        filename = state
        self.__init__(filename)

    @property
    def val(self):
        """The NumPy memmap array of floats that represents the values.  You can get this property, but cannot set it (except with itself)

        >>> from pysnptools.pstreader import PstMemMap
        >>> from pysnptools.util import example_file # Download and return local file name
        >>>
        >>> pst_mem_map_file = example_file('pysnptools/examples/tiny.pst.memmap')
        >>> pst_mem_map = PstMemMap(pst_mem_map_file)
        >>> print(pst_mem_map.val[0,1])
        2.0
        """
        self._run_once()
        return self._val

    @val.setter
    def val(self, new_value):
        self._run_once()
        if self._val is new_value:
            return
        raise Exception("PstMemMap val's cannot be set to a different array")

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def row_property(self):
        self._run_once()
        return self._row_property

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    @property
    def offset(self):
        """The byte position in the file where the memory-mapped values start."""
        self._run_once()
        return self._offset

    @property
    def filename(self):
        """The name of the memory-mapped file"""
        # Don't need '_run_once'
        return self._filename

    @staticmethod
    def empty(
        row,
        col,
        filename,
        row_property=None,
        col_property=None,
        order="F",
        dtype=np.float64,
        val_shape=None,
    ):
        """Create an empty :class:`.PstMemMap` on disk.

        :param row: The :attr:`PstReader.row` information
        :type row: an array of anything

        :param col: The :attr:`PstReader.col` information
        :type col: an array of anything

        :param filename: name of memory-mapped file to create
        :type filename: string

        :param row_property: optional -- The additional :attr:`PstReader.row_property` information associated with each row. Default: None
        :type row_property: an array of anything

        :param col_property: optional -- The additional :attr:`PstReader.col_property` information associated with each col. Default: None
        :type col_property: an array of anything

        :param order: {'F' (default), 'C'}, optional -- Specify the order of the ndarray.
        :type order: string or None

        :param dtype: {numpy.float64 (default), numpy.float32}, optional -- The data-type for the :attr:`PstMemMap.val` ndarray.
        :type dtype: data-type

        :param val_shape: (Default: None), optional -- The shape of the last dimension of :attr:`PstMemMap.val`. *None* means each value is a scalar.
        :type val_shape: None or a number

        :rtype: :class:`.PstMemMap`

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.pstreader import PstMemMap
        >>> filename = "tempdir/tiny.pst.memmap"
        >>> pstutil.create_directory_if_necessary(filename)
        >>> pst_mem_map = PstMemMap.empty(row=['a','b','c'],col=['y','z'],filename=filename,row_property=['A','B','C'],order="F",dtype=np.float64)
        >>> pst_mem_map.val[:,:] = [[1,2],[3,4],[np.nan,6]]
        >>> pst_mem_map.flush()

        """
        dtype = np.dtype(dtype)
        self = PstMemMap(filename)
        self._empty_inner(
            row, col, filename, row_property, col_property, order, dtype, val_shape
        )
        return self

    def _empty_inner(
        self, row, col, filename, row_property, col_property, order, dtype, val_shape
    ):
        self._ran_once = True
        self._dtype = np.dtype(dtype)
        self._order = order

        row = PstData._fixup_input(row)
        col = PstData._fixup_input(col)
        row_property = PstData._fixup_input(row_property, count=len(row))
        col_property = PstData._fixup_input(col_property, count=len(col))

        with open(filename, "wb") as fp:
            np.save(fp, np.array([_magic_number]))
            np.save(fp, np.array(["pstmemmap"]))  # name of file format
            np.save(fp, np.array([2]))  # file format version
            np.save(fp, row)
            np.save(fp, col)
            np.save(fp, row_property)
            np.save(fp, col_property)
            np.save(fp, np.array([self._dtype]))
            np.save(fp, np.array([self._order]))
            np.save(fp, np.array([val_shape]))
            self._offset = fp.tell()

        logging.info("About to start allocating memmap '{0}'".format(filename))
        shape = (
            (len(row), len(col))
            if val_shape is None
            else (len(row), len(col), val_shape)
        )
        val = np.memmap(
            filename,
            offset=self._offset,
            dtype=dtype,
            mode="r+",
            order=order,
            shape=shape,
        )
        logging.info(
            "Finished allocating memmap '{0}'. Size is {1:,}".format(
                filename, os.path.getsize(filename)
            )
        )
        PstData.__init__(
            self,
            row,
            col,
            val,
            row_property,
            col_property,
            name="np.memmap('{0}')".format(filename),
        )

    def _run_once(self):
        if self._ran_once:
            return
        row, col, val, row_property, col_property = self._run_once_inner()
        PstData.__init__(
            self,
            row,
            col,
            val,
            row_property,
            col_property,
            name="np.memmap('{0}')".format(self._filename),
        )

    def _run_once_inner(self):
        self._ran_once = True

        logging.debug("np.load('{0}')".format(self._filename))
        with open(self._filename, "rb") as fp:
            first = np.load(fp, allow_pickle=True)
            if len(first) == 1 and first[0] == _magic_number:
                format = np.load(fp, allow_pickle=True)[0]
                version = np.load(fp, allow_pickle=True)[0]
                assert format == "pstmemmap", "Expect format of 'pstmemmap'"
                assert version == 2, "Expect version of 2"
                row = np.load(fp, allow_pickle=True)
                col = np.load(fp, allow_pickle=True)
                row_property = np.load(fp, allow_pickle=True)
                col_property = np.load(fp, allow_pickle=True)
                self._dtype = np.load(fp, allow_pickle=True)[0]
                self._order = np.load(fp, allow_pickle=True)[0]
                val_shape = np.load(fp, allow_pickle=True)[0]
                self._offset = fp.tell()
            else:  # Try to load version one
                row = first
                col = np.load(fp, allow_pickle=True)
                row_property = np.load(fp, allow_pickle=True)
                col_property = np.load(fp, allow_pickle=True)
                self._dtype = np.load(fp, allow_pickle=True)[0]
                self._order = np.load(fp, allow_pickle=True)[0]
                val_shape = None
                self._offset = fp.tell()

        shape = (
            (len(row), len(col))
            if val_shape is None
            else (len(row), len(col), val_shape)
        )
        val = np.memmap(
            self._filename,
            offset=self._offset,
            dtype=self._dtype,
            mode="r",
            order=self._order,
            shape=shape,
        )
        return row, col, val, row_property, col_property

    def copyinputs(self, copier):
        # doesn't need to self._run_once()
        copier.input(self._filename)

    # Most _read's support only indexlists or None, but this one supports Slices, too.
    _read_accepts_slices = True

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        dtype = np.dtype(dtype)
        force_python_only = True  # Memmap arrays may not be aligned to Rust's standards, so process via Python
        val, shares_memory = self._apply_sparray_or_slice_to_val(
            self.val,
            row_index_or_none,
            col_index_or_none,
            order,
            dtype,
            force_python_only,
            num_threads,
        )
        if shares_memory and not view_ok:
            val = val.copy(order="K")
        return val

    @staticmethod
    def _order(pstdata):
        if pstdata.val.flags["F_CONTIGUOUS"]:
            return "F"
        if pstdata.val.flags["C_CONTIGUOUS"]:
            return "C"
        raise Exception("Don't know order of PstData's value")

    def flush(self):
        """Flush :attr:`PstMemMap.val` to disk and close the file. (If values or properties are accessed again, the file will be reopened.)

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.pstreader import PstMemMap
        >>> filename = "tempdir/tiny.pst.memmap"
        >>> pstutil.create_directory_if_necessary(filename)
        >>> pst_mem_map = PstMemMap.empty(row=['a','b','c'],col=['y','z'],filename=filename,row_property=['A','B','C'],order="F",dtype=np.float64)
        >>> pst_mem_map.val[:,:] = [[1,2],[3,4],[np.nan,6]]
        >>> pst_mem_map.flush()

        """
        if self._ran_once:
            self._val.flush()
            del self._val
            self._val = None
            self._ran_once = False

    @staticmethod
    def write(filename, pstdata):
        """Writes a :class:`PstData` to :class:`PstMemMap` format and returns the :class:`.PstMemMap`.

        :param filename: the name of the file to create
        :type filename: string
        :param pstdata: The in-memory data that should be written to disk.
        :type pstdata: :class:`PstData`
        :rtype: :class:`.PstMemMap`

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.pstreader import PstData, PstMemMap
        >>> data1 = PstData(row=['a','b','c'],col=['y','z'],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> pstutil.create_directory_if_necessary("tempdir/tiny.pst.memmap")
        >>> PstMemMap.write("tempdir/tiny.pst.memmap",data1)      # Write data1 in PstMemMap format
        PstMemMap('tempdir/tiny.pst.memmap')
        """

        self = PstMemMap.empty(
            pstdata.row,
            pstdata.col,
            filename + ".temp",
            row_property=pstdata.row_property,
            col_property=pstdata.col_property,
            order=PstMemMap._order(pstdata),
            dtype=pstdata.val.dtype,
            val_shape=pstdata.val_shape,
        )
        if pstdata.val_shape is None:
            self.val[:, :] = pstdata.val
        else:
            self.val[:, :, :] = pstdata.val
        self.flush()
        if os.path.exists(filename):
            os.remove(filename)
        shutil.move(filename + ".temp", filename)
        logging.debug("Done writing " + filename)

        return PstMemMap(filename)


class TestPstMemMap(unittest.TestCase):
    def test1(self):
        logging.info("in TestPstMemMap test1")
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        filename2 = "tempdir/tiny.pst.memmap"
        pstutil.create_directory_if_necessary(filename2)
        pstreader2 = PstMemMap.empty(
            row=["a", "b", "c"],
            col=["y", "z"],
            filename=filename2,
            row_property=["A", "B", "C"],
            order="F",
            dtype=np.float64,
        )
        assert isinstance(pstreader2.val, np.memmap)
        pstreader2.val[:, :] = [[1, 2], [3, 4], [np.nan, 6]]
        assert np.array_equal(
            pstreader2[[0], [0]].read(view_ok=True).val, np.array([[1.0]])
        )
        pstreader2.flush()
        assert isinstance(pstreader2.val, np.memmap)
        assert np.array_equal(
            pstreader2[[0], [0]].read(view_ok=True).val, np.array([[1.0]])
        )
        pstreader2.flush()

        pstreader3 = PstMemMap(filename2)
        assert np.array_equal(
            pstreader3[[0], [0]].read(view_ok=True).val, np.array([[1.0]])
        )
        assert isinstance(pstreader3.val, np.memmap)

        pstreader = PstMemMap("../examples/tiny.pst.memmap")
        assert pstreader.row_count == 3
        assert pstreader.col_count == 2
        assert isinstance(pstreader.val, np.memmap)

        pstdata = pstreader.read(view_ok=True)
        assert isinstance(pstdata.val, np.memmap)
        os.chdir(old_dir)

    def test2(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        filename = "tempdir/x.pst.memmap"
        pstutil.create_directory_if_necessary(filename)

        a = PstMemMap.empty(
            row=["a", "b", "c"],
            col=["y", "z"],
            filename=filename,
            row_property=["A", "B", "C"],
            order="F",
            dtype=np.float64,
        )
        b = PstData(
            row=["a", "b", "c"],
            col=["y", "z"],
            val=[[1, 2], [3, 4], [np.nan, 6]],
            row_property=["A", "B", "C"],
        )
        pointer1, read_only_flag = a.val.__array_interface__["data"]
        a.val += 1
        a.val += b.val
        pointer2, read_only_flag = a.val.__array_interface__["data"]
        assert pointer1 == pointer2
        os.chdir(old_dir)

    def test3(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        filename = "tempdir/x.pst.memmap"
        pstutil.create_directory_if_necessary(filename)

        a = PstMemMap.empty(
            row=["a", "b", "c"],
            col=["y", "z"],
            filename=filename,
            row_property=["A", "B", "C"],
            order="F",
            dtype=np.float64,
        )
        pstdata = a.read(order="C", view_ok=True)
        os.chdir(old_dir)


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPstMemMap))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    if True:
        from pysnptools.pstreader import PstMemMap

        fn = "../examples/tiny.pst.memmap"
        os.getcwd()
        print((os.path.exists(fn)))
        pst_mem_map = PstMemMap(fn)
        print((pst_mem_map.val[0, 1]))

    if False:
        a = np.ndarray([2, 3])
        pointer, read_only_flag = a.__array_interface__["data"]
        print(pointer)
        a *= 2
        pointer, read_only_flag = a.__array_interface__["data"]
        print(pointer)
        a = PstMemMap.empty(
            row=["a", "b", "c"],
            col=["y", "z"],
            filename=r"c:\deldir\a.memmap",
            row_property=["A", "B", "C"],
            order="F",
            dtype=np.float64,
        )
        b = PstData(
            row=["a", "b", "c"],
            col=["y", "z"],
            val=[[1, 2], [3, 4], [np.nan, 6]],
            row_property=["A", "B", "C"],
        )
        pointer, read_only_flag = a.val.__array_interface__["data"]
        print(pointer)
        a.val += 1
        a.val += b.val
        pointer, read_only_flag = a.val.__array_interface__["data"]
        print(pointer)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    result = doctest.testmod(optionflags=doctest.ELLIPSIS)
    assert result.failed == 0, "failed doc test: " + __file__



import numpy as np
import subprocess, sys, os.path
from itertools import *
import pandas as pd
import logging
from pysnptools.snpreader import SnpReader
from pysnptools.snpreader import SnpData
import warnings
from pysnptools.pstreader import _OneShot


# LATER say that this is also called dosage or bimbam
# LATER work when the extension is \*.dosage or anything else
# LATER make so fam can be called bam and map can be called bim
class Dat(_OneShot, SnpReader):
    """
    A :class:`.SnpReader` for reading Dat/Fam/Map-formatted files from disk.

    See :class:`.SnpReader` for general examples of using SnpReaders.

    This is a text format that can store any numeric values. (In contrast, Bed and Ped can only store 0,1,2, and missing). Its Dat files look like::

        null_0  	j	n	0.333	1	2
        null_100	j	n	2	1	1
        null_200	j	n	0	nan	1
        ...

    Its Map and Fam files are described `here <http://zzz.bwh.harvard.edu/plink/data.shtml>`__.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The Dat file to read.
                     * **skiprows** (*int*) -- Number of lines to skip before reading. Defaults to 0.

        :Example:

        >>> from pysnptools.snpreader import Dat
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> dat_file = example_file("pysnptools/examples/toydata.*","*.dat")
        >>> data_on_disk = Dat(dat_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 10000)

    **Methods beyond** :class:`.SnpReader`
    """

    def __init__(self, filename, skiprows=0):
        """
        filename    : string of the name of the Dat file.
        """
        super(Dat, self).__init__()
        self.filename = SnpReader._name_of_other_file(
            filename, remove_suffix="dat", add_suffix="dat"
        )
        self.skiprows = skiprows

    def _read_pstdata(self):
        row = SnpReader._read_fam(self.filename, remove_suffix="dat")
        col, col_property = SnpReader._read_map_or_bim(
            self.filename, remove_suffix="dat", add_suffix="map"
        )
        if len(row) == 0 or len(col) == 0:
            return SnpData(
                iid=row, sid=col, pos=col_property, val=np.empty([len(row), len(col)])
            )
        datfields = pd.read_csv(
            self.filename,
            delimiter="\t",
            header=None,
            index_col=False,
            skiprows=self.skiprows,
        )
        if not np.array_equal(datfields[0], col):
            raise Exception(
                "Expect snp list in map file to exactly match snp list in dat file"
            )
        del datfields[0]
        del datfields[1]
        del datfields[2]
        assert (
            len(row) == datfields.shape[1]
        ), "Expect # iids in fam file to match dat file"
        val = datfields.values.T
        snpdata = SnpData(iid=row, sid=col, pos=col_property, val=val)
        return snpdata

    def copyinputs(self, copier):
        # doesn't need to self.run_once() because creates name of all files itself
        copier.input(
            SnpReader._name_of_other_file(
                self.filename, remove_suffix="dat", add_suffix="dat"
            )
        )
        copier.input(
            SnpReader._name_of_other_file(
                self.filename, remove_suffix="dat", add_suffix="fam"
            )
        )
        copier.input(
            SnpReader._name_of_other_file(
                self.filename, remove_suffix="dat", add_suffix="map"
            )
        )

    @staticmethod
    def write(filename, snpdata):
        """Writes a :class:`SnpData` to dat/fam/map format and returns the :class:`.Dat`.

        :param filename: the name of the file to create
        :type filename: string
        :param snpdata: The in-memory data that should be written to disk.
        :type snpdata: :class:`SnpData`
        :rtype: :class:`.Dat`

        >>> from pysnptools.snpreader import Dat, Bed
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> snpdata = Bed(bed_file,count_A1=False)[:,:10].read()  # Read first 10 snps from Bed format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.dat")
        >>> Dat.write("tempdir/toydata10.dat",snpdata)              # Write data in dat/fam/map format
        Dat('tempdir/toydata10.dat')
        """

        if isinstance(filename, SnpData) and isinstance(
            snpdata, str
        ):  # For backwards compatibility, reverse inputs if necessary
            warnings.warn(
                "write statement should have filename before data to write",
                DeprecationWarning,
            )
            filename, snpdata = snpdata, filename

        SnpReader._write_fam(snpdata, filename, remove_suffix="dat")
        SnpReader._write_map_or_bim(
            snpdata, filename, remove_suffix="dat", add_suffix="map"
        )
        filename = SnpReader._name_of_other_file(
            filename, remove_suffix="dat", add_suffix="dat"
        )

        snpsarray = snpdata.val
        with open(filename, "w") as dat_filepointer:
            for sid_index, sid in enumerate(snpdata.sid):
                if sid_index % 1000 == 0:
                    logging.info(
                        "Writing snp # {0} to file '{1}'".format(sid_index, filename)
                    )
                dat_filepointer.write(
                    "%s\tj\tn\t" % sid
                )  # use "j" and "n" as the major and minor allele
                row = snpsarray[:, sid_index]
                dat_filepointer.write("\t".join((str(i) for i in row)) + "\n")
        logging.info("Done writing " + filename)
        return Dat(filename)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
